r"""
Cell BLAST based on DIRECTi models
"""

import collections
import glob
import os
import typing
import tempfile

import joblib
import numba
import numpy as np
import pandas as pd
import scipy.sparse
import scipy.stats
import sklearn.neighbors

from . import config, data, directi, metrics, utils

NORMAL = 1
MINIMAL = 0


def _wasserstein_distance_impl(x: np.ndarray, y: np.ndarray):  # pragma: no cover
    x_sorter = np.argsort(x)
    y_sorter = np.argsort(y)
    xy = np.concatenate((x, y))
    xy.sort()
    deltas = np.diff(xy)
    x_cdf = np.searchsorted(x[x_sorter], xy[:-1], 'right') / x.size
    y_cdf = np.searchsorted(y[y_sorter], xy[:-1], 'right') / y.size
    return np.sum(np.multiply(np.abs(x_cdf - y_cdf), deltas))


def _energy_distance_impl(x: np.ndarray, y: np.ndarray):  # pragma: no cover
    x_sorter = np.argsort(x)
    y_sorter = np.argsort(y)
    xy = np.concatenate((x, y))
    xy.sort()
    deltas = np.diff(xy)
    x_cdf = np.searchsorted(x[x_sorter], xy[:-1], "right") / x.size
    y_cdf = np.searchsorted(y[y_sorter], xy[:-1], "right") / y.size
    return np.sqrt(2 * np.sum(np.multiply(np.square(x_cdf - y_cdf), deltas)))


@numba.extending.overload(
    scipy.stats.wasserstein_distance, jit_options={"nogil": True, "cache": True})
def _wasserstein_distance(x: np.ndarray, y: np.ndarray):  # pragma: no cover
    if x == numba.float32[::1] and y == numba.float32[::1]:
        return _wasserstein_distance_impl


@numba.extending.overload(
    scipy.stats.energy_distance, jit_options={"nogil": True, "cache": True})
def _energy_distance(x: np.ndarray, y: np.ndarray):  # pragma: no cover
    if x == numba.float32[::1] and y == numba.float32[::1]:
        return _energy_distance_impl


@numba.jit(nopython=True, nogil=True, cache=True)
def ed(x: np.ndarray, y: np.ndarray):  # pragma: no cover
    r"""
    x : latent_dim
    y : latent_dim
    """
    return np.sqrt(np.square(x - y).sum())


@numba.jit(nopython=True, nogil=True, cache=True)
def _md(
        x: np.ndarray, y: np.ndarray, x_posterior: np.ndarray
) -> np.ndarray:  # pragma: no cover
    r"""
    x : latent_dim
    y : latent_dim
    x_posterior : n_posterior * latent_dim
    """
    if np.all(x == y):
        return 0.0
    x_posterior = x_posterior - x
    cov_x = np.dot(x_posterior.T, x_posterior) / x_posterior.shape[0]
    dev = np.expand_dims(y - x, axis=1)
    return np.sqrt(np.dot(dev.T, np.dot(np.linalg.inv(cov_x), dev)))[0, 0]


@numba.jit(nopython=True, nogil=True, cache=True)
def md(
        x: np.ndarray, y: np.ndarray,
        x_posterior: np.ndarray, y_posterior: np.ndarray
) -> np.ndarray:  # pragma: no cover
    r"""
    x : latent_dim
    y : latent_dim
    x_posterior : n_posterior * latent_dim
    y_posterior : n_posterior * latent_dim
    """
    if np.all(x == y):
        return 0.0
    return 0.5 * (_md(x, y, x_posterior) + _md(y, x, y_posterior))


@numba.jit(nopython=True, nogil=True, cache=True)
def _compute_pcasd(
        x: np.ndarray, x_posterior: np.ndarray, eps: float
) -> np.ndarray:  # pragma: no cover
    r"""
    x : latent_dim
    x_posterior : n_posterior * latent_dim
    """
    centered_x_posterior = x_posterior - np.sum(x_posterior, axis=0) / x_posterior.shape[0]
    cov_x = np.dot(centered_x_posterior.T, centered_x_posterior)
    v = np.real(np.linalg.eig(cov_x.astype(np.complex64))[1])  # Suppress domain change due to rounding errors
    x_posterior = np.dot(x_posterior - x, v)
    squared_x_posterior = np.square(x_posterior)
    asd = np.empty((2, x_posterior.shape[1]), dtype=np.float32)
    for p in range(x_posterior.shape[1]):
        mask = x_posterior[:, p] < 0
        asd[0, p] = np.sqrt((
            np.sum(squared_x_posterior[mask, p])
        ) / max(np.sum(mask), 1)) + eps
        asd[1, p] = np.sqrt((
            np.sum(squared_x_posterior[~mask, p])
        ) / max(np.sum(~mask), 1)) + eps
    return np.concatenate((v, asd), axis=0)


@numba.jit(nopython=True, nogil=True, cache=True)
def _compute_pcasd_across_models(
        x: np.ndarray, x_posterior: np.ndarray, eps: float = 1e-1
) -> np.ndarray:  # pragma: no cover
    r"""
    x : n_models * latent_dim
    x_posterior : n_models * n_posterior * latent_dim
    """
    result = np.empty((x.shape[0], x.shape[-1] + 2, x.shape[-1]), dtype=np.float32)
    for i in range(x.shape[0]):
        result[i] = _compute_pcasd(x[i], x_posterior[i], eps)
    return result


@numba.jit(nopython=True, nogil=True, cache=True)
def _amd(
        x: np.ndarray, y: np.ndarray, x_posterior: np.ndarray,
        eps: float, x_is_pcasd: bool = False
) -> np.ndarray:  # pragma: no cover
    r"""
    x : latent_dim
    y : latent_dim
    x_posterior : n_posterior * latent_dim
    """
    if np.all(x == y):
        return 0.0
    if not x_is_pcasd:
        x_posterior = _compute_pcasd(x, x_posterior, eps)
    v = x_posterior[:-2]
    asd = x_posterior[-2:]
    y = np.dot(y - x, v)
    for p in range(y.size):
        if y[p] < 0:
            y[p] /= asd[0, p]
        else:
            y[p] /= asd[1, p]
    return np.linalg.norm(y)


@numba.jit(nopython=True, nogil=True, cache=True)
def amd(
        x: np.ndarray, y: np.ndarray,
        x_posterior: np.ndarray, y_posterior: np.ndarray, eps: float = 1e-1,
        x_is_pcasd: bool = False, y_is_pcasd: bool = False
) -> np.ndarray:  # pragma: no cover
    r"""
    x : latent_dim
    y : latent_dim
    x_posterior : n_posterior * latent_dim
    y_posterior : n_posterior * latent_dim
    """
    if np.all(x == y):
        return 0.0
    return 0.5 * (
        _amd(x, y, x_posterior, eps, x_is_pcasd) +
        _amd(y, x, y_posterior, eps, y_is_pcasd)
    )


@numba.jit(nopython=True, nogil=True, cache=True)
def npd_v1(
        x: np.ndarray, y: np.ndarray,
        x_posterior: np.ndarray, y_posterior: np.ndarray, eps: float = 0.0
) -> np.ndarray:  # pragma: no cover
    r"""
    x : latent_dim
    y : latent_dim
    x_posterior : n_posterior * latent_dim
    y_posterior : n_posterior * latent_dim
    """
    projection = x - y  # latent_dim
    if np.all(projection == 0.0):
        projection[...] = 1.0  # any projection is equivalent
    projection /= np.linalg.norm(projection)
    x_posterior = np.sum(x_posterior * projection, axis=1)  # n_posterior_samples
    y_posterior = np.sum(y_posterior * projection, axis=1)  # n_posterior_samples
    xy_posterior = np.concatenate((x_posterior, y_posterior))
    xy_posterior1 = (xy_posterior - np.mean(x_posterior)) / (np.std(x_posterior) + np.float32(eps))
    xy_posterior2 = (xy_posterior - np.mean(y_posterior)) / (np.std(y_posterior) + np.float32(eps))
    return 0.5 * (scipy.stats.wasserstein_distance(
        xy_posterior1[:len(x_posterior)],
        xy_posterior1[-len(y_posterior):]
    ) + scipy.stats.wasserstein_distance(
        xy_posterior2[:len(x_posterior)],
        xy_posterior2[-len(y_posterior):]
    ))


@numba.jit(nopython=True, nogil=True, cache=True)
def _npd_v2(
        x: np.ndarray, y: np.ndarray, x_posterior: np.ndarray, eps: float
) -> np.ndarray:  # pragma: no cover
    r"""
    x : latent_dim
    y : latent_dim
    x_posterior : n_posterior * latent_dim
    """
    if np.all(x == y):
        return 0.0
    dev = y - x
    udev = dev / np.linalg.norm(dev)
    projected_noise = np.sum((x_posterior - x) * udev, axis=1)
    projected_y = np.sum((y - x) * udev)
    mask = (projected_noise * projected_y) >= 0
    scaler = np.sqrt(
        np.sum(np.square(projected_noise[mask])) /
        max(np.sum(mask), 1)
    )
    return np.abs(projected_y) / (scaler + eps)


@numba.jit(nopython=True, nogil=True, cache=True)
def npd_v2(
        x: np.ndarray, y: np.ndarray,
        x_posterior: np.ndarray, y_posterior: np.ndarray, eps: float = 1e-1
) -> np.ndarray:  # pragma: no cover
    r"""
    x : latent_dim
    y : latent_dim
    x_posterior : n_posterior * latent_dim
    y_posterior : n_posterior * latent_dim
    """
    if np.all(x == y):
        return 0.0
    return 0.5 * (
        _npd_v2(x, y, x_posterior, eps) +
        _npd_v2(y, x, y_posterior, eps)
    )


@numba.jit(nopython=True, nogil=True, cache=True)
def _hit_ed_across_models(
        query_latent: np.ndarray,
        ref_latent: np.ndarray
) -> np.ndarray:  # pragma: no cover
    r"""
    query_latent : n_models * latent_dim
    ref_latent : n_hits * n_models * latent_dim
    returns : n_hits * n_models
    """
    dist = np.empty(ref_latent.shape[:-1])  # n_hits * n_models
    for i in range(dist.shape[1]):  # model index
        x = query_latent[i, ...]  # latent_dim
        for j in range(dist.shape[0]):  # hit index
            y = ref_latent[j, i, ...]  # latent_dim
            dist[j, i] = ed(x, y)
    return dist


@numba.jit(nopython=True, nogil=True, cache=True)
def _hit_md_across_models(
        query_latent: np.ndarray, ref_latent: np.ndarray,
        query_posterior: np.ndarray, ref_posterior: np.ndarray
) -> np.ndarray:  # pragma: no cover
    r"""
    query_latent : n_models * latent_dim
    ref_latent : n_hits * n_models * latent_dim
    query_posterior : n_models * n_posterior * latent_dim
    ref_posterior : n_hits * n_models * n_posterior * latent_dim
    returns : n_hits * n_models
    """
    dist = np.empty(ref_latent.shape[:-1])  # n_hits * n_models
    for i in range(dist.shape[1]):  # model index
        x = query_latent[i, ...]  # latent_dim
        x_posterior = query_posterior[i, ...]  # n_posterior * latent_dim
        for j in range(dist.shape[0]):  # hit index
            y = ref_latent[j, i, ...]  # latent_dim
            y_posterior = ref_posterior[j, i, ...]  # n_posterior * latent_dim
            dist[j, i] = md(x, y, x_posterior, y_posterior)
    return dist


@numba.jit(nopython=True, nogil=True, cache=True)
def _hit_amd_across_models(
        query_latent: np.ndarray, ref_latent: np.ndarray,
        query_posterior: np.ndarray, ref_posterior: np.ndarray, eps: float = 1e-1
) -> np.ndarray:  # pragma: no cover
    r"""
    query_latent : n_models * latent_dim
    ref_latent : n_hits * n_models * latent_dim
    query_posterior : n_models * n_posterior * latent_dim
    ref_posterior : n_hits * n_models * n_posterior * latent_dim
    returns : n_hits * n_models
    """
    dist = np.empty(ref_latent.shape[:-1])  # n_hits * n_models
    for i in range(dist.shape[1]):  # model index
        x = query_latent[i, ...]  # latent_dim
        x_posterior = query_posterior[i, ...]  # n_posterior * latent_dim
        for j in range(dist.shape[0]):  # hit index
            y = ref_latent[j, i, ...]  # latent_dim
            y_posterior = ref_posterior[j, i, ...]  # n_posterior * latent_dim
            dist[j, i] = amd(
                x, y, x_posterior, y_posterior,
                eps=eps, x_is_pcasd=False, y_is_pcasd=True
            )
    return dist


@numba.jit(nopython=True, nogil=True, cache=True)
def _hit_npd_v1_across_models(
        query_latent: np.ndarray, ref_latent: np.ndarray,
        query_posterior: np.ndarray, ref_posterior: np.ndarray,
        eps: float = 0.0
) -> np.ndarray:  # pragma: no cover
    r"""
    query_latent : n_models * latent_dim
    ref_latent : n_hits * n_models * latent_dim
    query_posterior : n_models * n_posterior * latent_dim
    ref_posterior : n_hits * n_models * n_posterior * latent_dim
    returns : n_hits * n_models
    """
    dist = np.empty(ref_latent.shape[:-1])  # n_hits * n_models
    for i in range(dist.shape[1]):  # model index
        x = query_latent[i, ...]  # latent_dim
        x_posterior = query_posterior[i, ...]  # n_posterior * latent_dim
        for j in range(dist.shape[0]):  # hit index
            y = ref_latent[j, i, ...]  # latent_dim
            y_posterior = ref_posterior[j, i, ...]  # n_posterior * latent_dim
            dist[j, i] = npd_v1(x, y, x_posterior, y_posterior, eps=eps)
    return dist


@numba.jit(nopython=True, nogil=True, cache=True)
def _hit_npd_v2_across_models(
        query_latent: np.ndarray, ref_latent: np.ndarray,
        query_posterior: np.ndarray, ref_posterior: np.ndarray,
        eps: float = 1e-1
) -> np.ndarray:  # pragma: no cover
    r"""
    query_latent : n_models * latent_dim
    ref_latent : n_hits * n_models * latent_dim
    query_posterior : n_models * n_posterior * latent_dim
    ref_posterior : n_hits * n_models * n_posterior * latent_dim
    returns : n_hits * n_models
    """
    dist = np.empty(ref_latent.shape[:-1])  # n_hits * n_models
    for i in range(dist.shape[1]):  # model index
        x = query_latent[i, ...]  # latent_dim
        x_posterior = query_posterior[i, ...]  # n_posterior * latent_dim
        for j in range(dist.shape[0]):  # hit index
            y = ref_latent[j, i, ...]  # latent_dim
            y_posterior = ref_posterior[j, i, ...]  # n_posterior * latent_dim
            dist[j, i] = npd_v2(x, y, x_posterior, y_posterior, eps=eps)
    return dist


DISTANCE_METRIC_ACROSS_MODELS = {
    ed: _hit_ed_across_models,
    md: _hit_md_across_models,
    amd: _hit_amd_across_models,
    npd_v1: _hit_npd_v1_across_models,
    npd_v2: _hit_npd_v2_across_models
}


class BLAST(object):

    r"""
    Cell BLAST

    Parameters
    ----------
    models
        A list of "DIRECTi" models.
    ref
        A reference dataset.
    distance_metric
        Cell-to-cell distance metric to use, should be among
        {"npd_v1", "npd_v2", "md", "amd", "ed"}.
    n_posterior
        How many samples from the posterior distribution to use for
        estimating posterior distance. Irrelevant for distance_metric="ed".
    n_empirical
        Number of random cell pairs to use when estimating empirical
        distribution of cell-to-cell distance.
    cluster_empirical
        Whether to build an empirical distribution for each intrinsic cluster
        independently.
    eps
        A small number added to the normalization factors used in certain
        posterior-based distance metrics to improve numeric stability.
        If not specified, a recommended value will be used according to the
        specified distance metric.
    force_components
        Whether to compute all the necessary components upon initialization.
        If set to False, necessary components will be computed
        on the fly when performing queries.

    Examples
    --------

    A typical BLAST pipeline is described below.

    Assuming we have a list of :class:`directi.DIRECTi` models already fitted
    on some reference data, we can construct a BLAST object by feeding the
    pretrained models and the reference data to the
    :class:`BLAST` constructor.

    >>> blast = BLAST(models, reference)

    We can efficiently query the reference and obtain initial hits via the
    :meth:`BLAST.query` method:

    >>> hits = blast.query(query)

    Then we filter the initial hits by using more accurate metrics
    e.g. empirical p-value based on posterior distance), and pooling together
    information across multiple models.

    >>> hits = hits.reconcile_models().filter(by="pval", cutoff=0.05)

    Finally, we use the :meth:`BLAST.annotate` method to obtain predictions
    based on reference annotations, e.g. "cell_ontology_class" in this case.

    >>> annotation = hits.annotate("cell_ontology_class")

    See the BLAST ipython notebook (:ref:`vignettes`) for live examples.
    """

    def __init__(
            self, models: typing.List[directi.DIRECTi],
            ref: data.ExprDataSet, distance_metric: str = "npd_v1",
            n_posterior: int = 50, n_empirical: int = 10000,
            cluster_empirical: bool = False, eps: typing.Optional[float] = None,
            force_components: bool = True, **kwargs
    ) -> None:
        self.models = models
        self.ref = data.ExprDataSet(
            ref.exprs, ref.obs.copy(), ref.var.copy(), ref.uns.copy()
        )  # exprs and uns are shallow copied, obs and var are deep copied

        self.latent = None
        self.nearest_neighbors = None
        self.cluster = None
        self.posterior = np.array([None] * self.ref.shape[0])
        self.empirical = None

        self.distance_metric = globals()[distance_metric] \
            if isinstance(distance_metric, str) else distance_metric
        self.n_posterior = n_posterior if self.distance_metric is not ed else 0
        self.n_empirical = n_empirical
        self.cluster_empirical = cluster_empirical
        self.eps = eps

        if force_components:
            self._force_components(**kwargs)

    def __len__(self) -> int:
        return len(self.models)

    def __getitem__(self, s) -> "BLAST":
        blast = BLAST(
            np.array(self.models)[s].tolist(),
            self.ref, self.distance_metric, self.n_posterior, self.n_empirical,
            self.cluster_empirical, self.eps, force_components=False
        )
        blast.latent = self.latent[:, s, ...] if self.latent is not None else None
        blast.cluster = self.cluster[:, s, ...] if self.cluster is not None else None
        blast.nearest_neighbors = np.array(self.nearest_neighbors)[s].tolist() \
            if self.nearest_neighbors is not None else None
        if self.posterior is not None:
            for i in range(self.posterior.size):
                if self.posterior[i] is not None:
                    blast.posterior[i] = self.posterior[i][s, ...]
        blast.empirical = [
            item for item in np.array(self.empirical)[s]
        ] if self.empirical is not None else None
        return blast

    def _get_latent(self, n_jobs: int) -> np.ndarray:  # n_cells * n_models * latent_dim
        if self.latent is None:
            utils.logger.info("Projecting to latent space...")
            self.latent = np.stack(joblib.Parallel(
                n_jobs=min(n_jobs, len(self)), backend="threading"
            )(joblib.delayed(model.inference)(
                self.ref
            ) for model in self.models), axis=1)
        return self.latent

    def _get_cluster(self, n_jobs: int) -> np.ndarray:  # n_cells * n_models
        if self.cluster is None:
            utils.logger.info("Obtaining intrinsic clustering...")
            self.cluster = np.stack(joblib.Parallel(
                n_jobs=min(n_jobs, len(self)), backend="threading"
            )(joblib.delayed(model.clustering)(
                self.ref
            ) for model in self.models), axis=1)
        return self.cluster

    def _get_posterior(
            self, n_jobs: int, random_seed: int,
            idx: typing.Optional[np.ndarray] = None
    ) -> np.ndarray:  # n_cells * (n_models * n_posterior * latent_dim)
        if idx is None:
            idx = np.arange(self.ref.shape[0])
        new_idx = np.intersect1d(np.unique(idx), np.where(np.vectorize(
            lambda x: x is None
        )(self.posterior))[0])
        if new_idx.size:
            utils.logger.info("Sampling from posteriors...")
            new_ref = self.ref[new_idx, :]
            new_posterior = np.stack(joblib.Parallel(
                n_jobs=min(n_jobs, len(self)), backend="threading"
            )(joblib.delayed(model.inference)(
                new_ref, n_posterior=self.n_posterior, random_seed=random_seed
            ) for model in self.models), axis=1)  # n_cells * n_models * n_posterior * latent_dim
            # NOTE: Slow discontigous memcopy here, but that's necessary since
            # we will be caching values by cells. It also makes values more
            # contiguous and faster to access in later cell-based operations.
            if self.distance_metric is amd:
                dist_kws = {"eps": self.eps} if self.eps is not None else {}
                new_latent = self._get_latent(n_jobs)[new_idx]
                self.posterior[new_idx] = joblib.Parallel(
                    n_jobs=n_jobs, backend="threading"
                )(joblib.delayed(_compute_pcasd_across_models)(
                    _new_latent, _new_posterior, **dist_kws
                ) for _new_latent, _new_posterior in zip(new_latent, new_posterior))
            else:
                self.posterior[new_idx] = [item for item in new_posterior]  # NOTE: No memcopy here
        return self.posterior[idx]

    def _get_nearest_neighbors(
            self, n_jobs: int
    ) -> typing.List[sklearn.neighbors.NearestNeighbors]:  # n_models
        if self.nearest_neighbors is None:
            latent = self._get_latent(n_jobs).swapaxes(0, 1)
            # NOTE: Makes cells discontiguous, but for nearest neighbor tree,
            # there's no influence on performance
            utils.logger.info("Fitting nearest neighbor trees...")
            self.nearest_neighbors = joblib.Parallel(
                n_jobs=min(n_jobs, len(self)), backend="loky"
            )(joblib.delayed(self._fit_nearest_neighbors)(
                _latent
            ) for _latent in latent)
        return self.nearest_neighbors

    def _get_empirical(
            self, n_jobs: int, random_seed: int
    ) -> np.ndarray:  # n_models * [n_clusters * n_empirical]
        if self.empirical is None:
            utils.logger.info("Generating empirical null distributions...")
            if not self.cluster_empirical:
                self.cluster = np.zeros((
                    self.ref.shape[0], len(self)
                ), dtype=np.int)
            latent = self._get_latent(n_jobs)
            cluster = self._get_cluster(n_jobs)
            rs = np.random.RandomState(random_seed)
            bg = rs.choice(latent.shape[0], size=self.n_empirical)
            if self.distance_metric is not ed:
                bg_posterior = self._get_posterior(n_jobs, random_seed, idx=bg)
            self.empirical = []
            dist_kws = {"eps": self.eps} if self.eps is not None else {}
            if self.distance_metric is amd:
                dist_kws["x_is_pcasd"] = True
                dist_kws["y_is_pcasd"] = True
            for k in range(len(self)):  # model_idx
                empirical = np.zeros((np.max(cluster[:, k]) + 1, self.n_empirical))
                for c in np.unique(cluster[:, k]):  # cluster_idx
                    fg = rs.choice(np.where(cluster[:, k] == c)[0], size=self.n_empirical)
                    if self.distance_metric is ed:
                        empirical[c] = np.sort(joblib.Parallel(
                            n_jobs=n_jobs, backend="threading"
                        )(joblib.delayed(self.distance_metric)(
                            latent[fg[i]], latent[bg[i]]
                        ) for i in range(self.n_empirical)))
                    else:
                        fg_posterior = self._get_posterior(n_jobs, random_seed, idx=fg)
                        empirical[c] = np.sort(joblib.Parallel(
                            n_jobs=n_jobs, backend="threading"
                        )(joblib.delayed(self.distance_metric)(
                            latent[fg[i], k], latent[bg[i], k],
                            fg_posterior[i][k], bg_posterior[i][k],
                            **dist_kws
                        ) for i in range(self.n_empirical)))
                self.empirical.append(empirical)
        return self.empirical

    def _force_components(
            self, n_jobs: int = config._USE_GLOBAL,
            random_seed: int = config._USE_GLOBAL
    ) -> None:
        n_jobs = config.N_JOBS if n_jobs == config._USE_GLOBAL else n_jobs
        random_seed = config.RANDOM_SEED if random_seed == config._USE_GLOBAL else random_seed
        self._get_nearest_neighbors(n_jobs)
        if self.distance_metric is not ed:
            self._get_posterior(n_jobs, random_seed)
        self._get_empirical(n_jobs, random_seed)

    @staticmethod
    def _fit_nearest_neighbors(
            x: np.ndarray
    ) -> sklearn.neighbors.NearestNeighbors:
        return sklearn.neighbors.NearestNeighbors().fit(x)

    @staticmethod
    def _nearest_neighbor_search(
            nn: sklearn.neighbors.NearestNeighbors,
            query: np.ndarray, n_neighbors: int
    ) -> np.ndarray:
        return nn.kneighbors(query, n_neighbors=n_neighbors)[1]

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def _nearest_neighbor_merge(x: np.ndarray) -> np.ndarray:  # pragma: no cover
        return np.unique(x)

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def _empirical_pvalue(
            hits: np.ndarray, dist: np.ndarray,
            cluster: np.ndarray, empirical: np.ndarray
    ) -> np.ndarray:  # pragma: no cover
        r"""
        hits : n_hits
        dist : n_hits * n_models
        cluster : n_cells * n_models
        empirical : n_models * [n_cluster * n_empirical]
        """
        pval = np.empty(dist.shape)
        for i in range(dist.shape[1]):  # model index
            for j in range(dist.shape[0]):  # hit index
                pval[j, i] = np.searchsorted(
                    empirical[i][cluster[hits[j], i]], dist[j, i]
                ) / empirical[i].shape[1]
        return pval

    def save(self, path: str, only_used_genes: bool = True) -> None:
        r"""
        Save BLAST object to a directory.

        Parameters
        ----------
        path
            Specifies a path to save the BLAST object.
        only_used_genes
            Whether to preserve only the genes used by models.
        """
        if not os.path.exists(path):
            os.makedirs(path)
        if self.ref is not None:
            if only_used_genes:
                if "__libsize__" not in self.ref.obs.columns:
                    self.ref.obs["__libsize__"] = np.array(
                        self.ref.exprs.sum(axis=1)
                    ).ravel()
                    # So that align will still work properly
                ref = self.ref[:, np.unique(np.concatenate([
                    model.genes for model in self.models
                ]))]
            ref.uns["distance_metric"] = self.distance_metric.__name__
            ref.uns["n_posterior"] = self.n_posterior
            ref.uns["n_empirical"] = self.n_empirical
            ref.uns["cluster_empirical"] = self.cluster_empirical
            ref.uns["eps"] = self.eps if self.eps is not None else "None"
            if self.latent is not None:
                ref.uns["latent"] = self.latent
            if self.latent is not None:
                ref.uns["cluster"] = self.cluster
            if self.empirical is not None:
                ref.uns["empirical"] = {str(i): item for i, item in enumerate(self.empirical)}
            ref.uns["posterior"] = {str(i): item for i, item in enumerate(self.posterior) if item is not None}
            ref.write_dataset(os.path.join(path, "ref.h5"))
        for i in range(len(self)):
            self.models[i].save(os.path.join(path, f"model_{i}"))

    @classmethod
    def load(cls, path: str, mode: int = NORMAL, **kwargs):
        r"""
        Load BLAST object from a directory.

        Parameters
        ----------
        path
            Specifies a path to load from.
        mode
            If mode is set to MINIMAL, model loading will be accelerated by only
            loading the encoders, but aligning BLAST (fine-tuning) would not be
            available. Should be among {cb.blast.NORMAL, cb.blast.MINIMAL}

        Returns
        -------
        blast
            Loaded BLAST object.
        """
        assert mode in (NORMAL, MINIMAL)
        ref = data.ExprDataSet.read_dataset(os.path.join(path, "ref.h5"))
        models = []
        for model_path in sorted(glob.glob(os.path.join(path, "model_*"))):
            models.append(directi.DIRECTi.load(model_path, _mode=mode))
        blast = cls(
            models, ref, ref.uns["distance_metric"], ref.uns["n_posterior"],
            ref.uns["n_empirical"], ref.uns["cluster_empirical"],
            None if ref.uns["eps"] == "None" else ref.uns["eps"],
            force_components=False
        )
        blast.latent = blast.ref.uns["latent"] if "latent" in blast.ref.uns else None
        blast.cluster = blast.ref.uns["cluster"] if "latent" in blast.ref.uns else None
        blast.empirical = [
            blast.ref.uns["empirical"][str(i)] for i in range(len(blast))
        ] if "empirical" in blast.ref.uns else None
        for i in range(ref.shape[0]):
            if str(i) in blast.ref.uns["posterior"]:
                blast.posterior[i] = blast.ref.uns["posterior"][str(i)]
        blast._force_components(**kwargs)
        return blast

    def query(
            self, query: data.ExprDataSet, n_neighbors: int = 5,
            store_dataset: bool = False,
            n_jobs: int = config._USE_GLOBAL,
            random_seed: int = config._USE_GLOBAL
    ) -> "Hits":
        r"""
        BLAST query

        Parameters
        ----------
        query
            Query transcriptomes.
        n_neighbors
            Initial number of nearest neighbors to search in each model.
        store_dataset
            Whether to store query dataset in the returned hit object.
            Note that this is necessary if :meth:`Hits.gene_deviation`
            is to be used.
        n_jobs
            Number of parallel jobs to run when performing query. If not
            specified, :data:`config.N_JOBS` will be used.
            Note that each (tensorflow) job could be distributed on multiple
            CPUs for a single "job".
        random_seed
            Random seed for posterior sampling. If not specified,
            :data:`utils.RANDOM_SEED` will be used.

        Returns
        -------
        hits
            Query hits
        """
        n_jobs = config.N_JOBS if n_jobs == config._USE_GLOBAL else n_jobs
        random_seed = config.RANDOM_SEED if random_seed == config._USE_GLOBAL else random_seed

        utils.logger.info("Projecting to latent space...")
        query_latent = joblib.Parallel(
            n_jobs=min(n_jobs, len(self)), backend="threading"
        )(joblib.delayed(model.inference)(
            query
        ) for model in self.models)  # n_models * [n_cells * latent_dim]

        utils.logger.info("Doing nearest neighbor search...")
        nearest_neighbors = self._get_nearest_neighbors(n_jobs)
        nni = np.stack(joblib.Parallel(
            n_jobs=min(n_jobs, len(self)), backend="threading"
        )(joblib.delayed(self._nearest_neighbor_search)(
            _nearest_neighbor, _query_latent, n_neighbors
        ) for _nearest_neighbor, _query_latent in zip(
            nearest_neighbors, query_latent
        )), axis=2)  # n_cells * n_neighbors * n_models

        utils.logger.info("Merging hits across models...")
        hits = joblib.Parallel(n_jobs=n_jobs, backend="threading")(
            joblib.delayed(self._nearest_neighbor_merge)(_nni) for _nni in nni
        )  # n_cells * [n_hits]
        hitsu, hitsi = np.unique(np.concatenate(hits), return_inverse=True)
        hitsi = np.split(hitsi, np.cumsum([item.size for item in hits])[:-1])

        query_latent = np.stack(query_latent, axis=1)  # n_cell * n_model * latent_dim
        ref_latent = self._get_latent(n_jobs)  # n_cell * n_model * latent_dim
        if self.distance_metric is ed:
            utils.logger.info("Computing Euclidean distances...")
            dist = joblib.Parallel(n_jobs=n_jobs, backend="threading")(
                joblib.delayed(_hit_ed_across_models)(
                    query_latent[i], ref_latent[hits[i]]
                ) for i in range(len(hits))
            )  # list of n_hits * n_models
        else:
            utils.logger.info("Computing posterior distribution distances...")
            query_posterior = np.stack(joblib.Parallel(
                n_jobs=min(n_jobs, len(self)), backend="threading"
            )(joblib.delayed(model.inference)(
                query, n_posterior=self.n_posterior, random_seed=random_seed
            ) for model in self.models), axis=1)  # n_cells * n_models * n_posterior_samples * latent_dim
            ref_posterior = np.stack(self._get_posterior(
                n_jobs, random_seed, idx=hitsu
            ))  # n_cells * n_models * n_posterior_samples * latent_dim
            distance_metric = DISTANCE_METRIC_ACROSS_MODELS[self.distance_metric]
            dist_kws = {"eps": self.eps} if self.eps is not None else {}
            dist = joblib.Parallel(n_jobs=n_jobs, backend="threading")(
                joblib.delayed(distance_metric)(
                    query_latent[i], ref_latent[hits[i]],
                    query_posterior[i], ref_posterior[hitsi[i]], **dist_kws
                ) for i in range(len(hits))
            )  # list of n_hits * n_models

        utils.logger.info("Computing empirical p-values...")
        empirical = self._get_empirical(n_jobs, random_seed)
        cluster = self._get_cluster(n_jobs)
        pval = joblib.Parallel(
            n_jobs=n_jobs, backend="threading"
        )(joblib.delayed(self._empirical_pvalue)(
            _hits, _dist, cluster, empirical
        ) for _hits, _dist in zip(hits, dist))  # list of n_hits * n_models

        return Hits(
            self, hits, dist, pval,
            query if store_dataset else data.ExprDataSet(
                scipy.sparse.csr_matrix((query.shape[0], 0)),
                obs=pd.DataFrame(index=query.obs.index),
                var=pd.DataFrame(), uns={}
            )
        )

    def align(
            self, query: typing.Union[
                data.ExprDataSet, typing.Mapping[str, data.ExprDataSet]
            ], n_jobs: int = config._USE_GLOBAL,
            random_seed: int = config._USE_GLOBAL,
            path: typing.Optional[str] = None, **kwargs
    ) -> "BLAST":
        r"""
        Align internal DIRECTi models with query datasets (fine tuning).

        Parameters
        ----------
        query
            A query dataset or a dict of query datasets, which will be aligned
            to the reference.
        n_jobs
            Number of parallel jobs to run when building the BLAST index,
            If not specified, :data:`config.N_JOBS` will be used.
            Note that each (tensorflow) job could be distributed on multiple
            CPUs for a single "job".
        random_seed
            Random seed for posterior sampling. If not specified,
            :data:`config.RANDOM_SEED` will be used.
        path
            Specifies a path to store temporary files.
        kwargs
            Additional keyword parameters passed to
            :meth:`directi.align_DIRECTi`.

        Returns
        -------
        blast
            A new BLAST object with aligned internal models.
        """
        if any(model._mode == directi._TEST for model in self.models):  # pragma: no cover
            raise Exception("Align not available!")
        n_jobs = config.N_JOBS if n_jobs == config._USE_GLOBAL else n_jobs
        random_seed = config.RANDOM_SEED if random_seed == config._USE_GLOBAL else random_seed
        path = path or tempfile.mkdtemp()

        aligned_models = joblib.Parallel(
            n_jobs=n_jobs, backend="threading"
        )(
            joblib.delayed(directi.align_DIRECTi)(
                self.models[i], self.ref, query, random_seed=random_seed,
                path=os.path.join(path, f"aligned_model_{i}"), **kwargs
            ) for i in range(len(self))
        )
        return BLAST(
            aligned_models, self.ref, distance_metric=self.distance_metric,
            n_posterior=self.n_posterior, n_empirical=self.n_empirical,
            cluster_empirical=self.cluster_empirical, eps=self.eps
        )


class Hits(object):

    r"""
    BLAST hits

    Parameters
    ----------
    ref
        The reference dataset.
    hits
        Indices of hit cell in the reference dataset.
        Each list element contains hit cell indices for a query cell.
    dist
        Hit cell distances.
        Each list element contains distances for a query cell.
        Each list element is a :math:`n\_hits \times n\_models` matrix,
        with matrix entries corresponding to the distance to
        each hit cell under each model.
    pval
        Hit cell empirical p-values.
        Each list element contains p-values for a query cell.
        Each list element is a :math:`n\_hits \times n\_models` matrix,
        with matrix entries corresponding to the empirical p-value of
        each hit cell under each model.
    names
        Query cell names.
    """

    FILTER_BY_DIST = 0
    FILTER_BY_PVAL = 1

    def __init__(
            self, blast: BLAST,
            hits: typing.List[np.ndarray],
            dist: typing.List[np.ndarray],
            pval: typing.List[np.ndarray],
            query: data.ExprDataSet
    ) -> None:
        self.blast = blast
        self.hits = np.array(hits)
        self.dist = np.array(dist)
        self.pval = np.array(pval)
        self.query = query
        if not self.hits.shape[0] == self.dist.shape[0] == \
                self.pval.shape[0] == self.query.shape[0]:
            raise ValueError("Inconsistent shape!")

    def __len__(self) -> int:
        return self.query.shape[0]

    def __iter__(self):
        for idx, (_hits, _dist, _pval) in enumerate(zip(self.hits, self.dist, self.pval)):
            yield Hits(self.blast, [_hits], [_dist], [_pval], self.query[idx, :])

    def __getitem__(self, s):
        s = [s] if isinstance(s, (int, np.integer)) else s
        return Hits(
            self.blast, self.hits[s], self.dist[s], self.pval[s],
            self.query[s, :]
        )

    def to_data_frames(self) -> typing.Mapping[str, pd.DataFrame]:
        r"""
        Construct hit data frames for query cells.
        Note that only reconciled ``Hits`` objects are supported.

        Returns
        -------
        data_frame_dicts
            Each element is hit data frame for a cell
        """
        if self.dist[0].ndim != 1 or self.pval[0].ndim != 1:
            raise RuntimeError("Please call `reconcile_models` first!")
        df_dict = collections.OrderedDict()
        for i, name in enumerate(self.query.obs_names):
            df_dict[name] = self.blast.ref.obs.iloc[self.hits[i], :].copy()
            df_dict[name].loc[:, "hits"] = self.hits[i]
            df_dict[name].loc[:, "dist"] = self.dist[i]
            df_dict[name].loc[:, "pval"] = self.pval[i]
        return df_dict

    def reconcile_models(
            self, dist_method: str = "mean", pval_method: str = "gmean"
    ) -> "Hits":
        r"""
        Integrate model-specific distances and empirical p-values.

        Parameters
        ----------
        dist_method
            Specifies how to integrate distances across difference models.
            Should be among {"mean", "gmean", "min", "max"}.
        pval_method
            Specifies how to integrate empirical p-values across different
            models. Should be among {"mean", "gmean", "min", "max"}.

        Returns
        -------
        reconciled_hits
            Hit object containing reconciled
        """
        dist_method = self._get_reconcile_method(dist_method)
        dist = [dist_method(item, axis=1) for item in self.dist]
        pval_method = self._get_reconcile_method(pval_method)
        pval = [pval_method(item, axis=1) for item in self.pval]
        return Hits(self.blast, self.hits, dist, pval, self.query)

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def _filter_hits(
            hits: np.ndarray, dist: np.ndarray, pval: np.ndarray,
            by: int, cutoff: float, model_tolerance: int
    ) -> typing.Tuple[np.ndarray, np.ndarray, np.ndarray]:  # pragma: no cover
        r"""
        hits : n_hits
        dist : n_hits * n_models
        pval : n_hits * n_models
        """
        if by == 0:  # Hits.FILTER_BY_DIST
            hit_mask = (dist.shape[1] - (dist <= cutoff).sum(axis=1)) <= model_tolerance
        else:  # Hits.FILTER_BY_PVAL
            hit_mask = (pval.shape[1] - (pval <= cutoff).sum(axis=1)) <= model_tolerance
        return hits[hit_mask], dist[hit_mask], pval[hit_mask]

    @staticmethod
    @numba.jit(nopython=True, nogil=True, cache=True)
    def _filter_reconciled_hits(
            hits: np.ndarray, dist: np.ndarray, pval: np.ndarray,
            by: int, cutoff: float
    ) -> typing.Tuple[np.ndarray, np.ndarray, np.ndarray]:  # pragma: no cover
        r"""
        hits : n_hits
        dist : n_hits
        pval : n_hits
        """
        if by == 0:  # Hits.FILTER_BY_DIST
            hit_mask = dist <= cutoff
        else:  # Hits.FILTER_BY_PVAL
            hit_mask = pval <= cutoff
        return hits[hit_mask], dist[hit_mask], pval[hit_mask]

    def filter(
            self, by: str = "pval", cutoff: float = 0.05,
            model_tolerance: int = 0, n_jobs: int = 1
    ) -> "Hits":
        r"""
        Filter hits by posterior distance or p-value

        Parameters
        ----------
        by
            Specifies a metric based on which to filter hits.
            Should be among {"dist", "pval"}.
        cutoff
            Cutoff when filtering hits.
        model_tolerance
            Maximal number of models allowed in which the cutoff is not
            satisfied, above which the query cell will be rejected.
            Irrelevant for reconciled hits.
        n_jobs
            Number of parallel jobs to run.

        Returns
        -------
        filtered_hits
            Hit object containing remaining hits after filtering
        """
        if by == "pval":
            assert self.pval is not None
            by = Hits.FILTER_BY_PVAL
        else:  # by == "dist"
            by = Hits.FILTER_BY_DIST
        if self.dist[0].ndim == 1:
            hits, dist, pval = [_ for _ in zip(*joblib.Parallel(
                n_jobs=n_jobs, backend="threading"
            )(joblib.delayed(self._filter_reconciled_hits)(
                _hits, _dist, _pval, by, cutoff
            ) for _hits, _dist, _pval in zip(
                self.hits, self.dist, self.pval
            )))]
        else:
            hits, dist, pval = [_ for _ in zip(*joblib.Parallel(
                n_jobs=n_jobs, backend="threading"
            )(joblib.delayed(self._filter_hits)(
                _hits, _dist, _pval, by, cutoff, model_tolerance
            ) for _hits, _dist, _pval in zip(
                self.hits, self.dist, self.pval
            )))]
        return Hits(self.blast, hits, dist, pval, self.query)

    def annotate(
            self, field: str, min_hits: int = 2,
            majority_threshold: float = 0.5, return_evidence: bool = False
    ) -> pd.DataFrame:
        r"""
        Annotate query cells based on existing annotations of hit cells
        via majority voting.
        Fields in the meta data or gene expression values can all be specified
        for annotation / prediction. Note that for gene expression, predicted
        values are in log-scale, and user should do proper normalization
        of the reference data in advance.

        Parameters
        ----------
        field
            Specifies a meta column or gene name to use for annotation.
        min_hits
            Minimal number of hits required for annotating a query cell,
            otherwise the query cell will be rejected.
        majority_threshold
            Minimal majority fraction (not inclusive) required for confident
            annotation. Only effective when predicting
            categorical variables. If the threshold is not met, annotation
            will be "ambiguous".
        return_evidence
            Whether to return evidence level of the annotations.

        Returns
        -------
        prediction
            Each row contains the inferred annotation for a query cell.
            If ``return_evidence`` is set to False, the data frame contains only
            one column, i.e. the inferred annotation.
            If ``return_evidence`` is set to True, the data frame also contains
            the number of hits, as well as the majority fraction (only for
            categorical annotations) for each query cell.
        """
        ref = self.blast.ref.get_meta_or_var(
            [field], normalize_var=False, log_var=True
        ).values.ravel()
        n_hits = np.repeat(0, len(self.hits))
        if np.issubdtype(ref.dtype.type, np.character) or \
                np.issubdtype(ref.dtype.type, np.object_):
            prediction = np.repeat("rejected", len(self.hits)).astype(object)
            majority_frac = np.repeat(np.nan, len(self.hits))
            for i, _hits in enumerate(self.hits):
                hits = ref[_hits]
                hits = hits[~utils.isnan(hits)] if hits.size else hits
                n_hits[i] = hits.size
                if n_hits[i] < min_hits:
                    continue
                label, count = np.unique(hits, return_counts=True)
                best_idx = count.argmax()
                majority_frac[i] = count[best_idx] / hits.size
                if majority_frac[i] <= majority_threshold:
                    prediction[i] = "ambiguous"
                    continue
                prediction[i] = label[best_idx]
            prediction = prediction.astype(ref.dtype.type)
        elif np.issubdtype(ref.dtype.type, np.number):
            prediction = np.repeat(np.nan, len(self.hits))
            for i, _hits in enumerate(self.hits):
                hits = ref[_hits]
                hits = hits[~utils.isnan(hits)] if hits.size else hits
                n_hits[i] = hits.size
                if n_hits[i] < min_hits:
                    continue
                prediction[i] = hits.mean()
                # np.array call is for 1-d mean that produces 0-d values
            prediction = np.stack(prediction, axis=0)
        else:  # pragma: no cover
            raise ValueError("Unsupported data type!")
        result = collections.OrderedDict()
        result[field] = prediction
        if return_evidence:
            result["n_hits"] = n_hits
            if "majority_frac" in locals():
                result["majority_frac"] = majority_frac
        return pd.DataFrame(result, index=self.query.obs_names)

    def blast2co(
            self, cl_dag: utils.CellTypeDAG,
            cl_field: str = "cell_ontology_class",
            min_hits: int = 2, thresh: float = 0.5, min_path: int = 4
    ) -> pd.DataFrame:
        r"""
        Annotate query cells based on existing annotations of hit cells
        via the cell-ontology-aware BLAST2CO method.

        Parameters
        ----------
        cl_dag
            Cell ontology DAG
        cl_field
            Specify the ``obs`` column containing cell ontology annotation
            in the reference dataset
        min_hits
            Minimal number of hits required for annotating a query cell,
            otherwise the query cell will be rejected.
        thresh
            Scoring threshold based on 1 - pvalue.
        min_path
            Minimal allowed value of the maximal distance to root
            for a prediction to be made.

        Returns
        -------
        prediction
            Each row contains the inferred annotation for a query cell.
        """
        if self.dist[0].ndim != 1 or self.pval[0].ndim != 1:
            raise RuntimeError("Please call `reconcile_models` first!")
        prediction = np.repeat("rejected", len(self.hits)).astype(object)
        ref = self.blast.ref.get_meta_or_var([cl_field]).values.ravel()
        for i, (_hits, _pval) in enumerate(zip(self.hits, self.pval)):
            hits = ref[_hits]
            if hits.size:
                mask = ~utils.isnan(hits)
                hits = hits[mask]
                _pval = _pval[mask]
            if hits.size < min_hits:
                continue
            cl_dag.value_reset()
            for cl in np.unique(hits):  # 1 - pval
                cl_dag.value_set(cl, np.sum(1 - _pval[hits == cl]) / np.sum(1 - _pval))
            cl_dag.value_update()
            leaves = cl_dag.best_leaves(
                thresh=thresh, min_path=min_path, retrieve=cl_field)
            if len(leaves) == 1:
                prediction[i] = leaves[0]
            elif len(leaves) > 1:
                prediction[i] = "ambiguous"
        return pd.DataFrame({cl_field: prediction}, index=self.query.obs_names)

    @staticmethod
    def _get_reconcile_method(method: str):
        if method == "mean":
            return np.mean
        if method == "gmean":
            return scipy.stats.gmean
        if method == "min":
            return np.min
        if method == "max":
            return np.max
        raise ValueError("Unknown method!")  # pragma: no cover

    def gene_gradient(
            self, eval_point: str = "query", normalize_deviation: bool = True,
            avg_models: bool = True, n_jobs: int = config._USE_GLOBAL
    ) -> typing.List[np.ndarray]:
        r"""
        Compute gene-wise gradient for each pair of query-hit cells
        based on query-hit deviation in the latent space. Useful for model
        interpretation.

        Parameters
        ----------
        eval_point
            At which point should the gradient be evaluated.
            Valid options include: {"query", "ref", "both"}
        normalize_deviation
            Whether to normalize query-hit deivation in the latent space.
        avg_models
            Whether to average gene-wise gradients across different models
        n_jobs
            Number of parallel jobs to run when performing query. If not
            specified, :data:`config.N_JOBS` will be used.
            Note that each (tensorflow) job could be distributed on multiple
            CPUs for a single "job".

        Returns
        -------
        gene_gradient
            A list with length equal to the number of query cells, where each
            element is a :class:`np.ndarray` containing gene-wise gradient for
            every hit cell of a query cell. The :class:`np.ndarray`s are of
            shape :math:`n\_hits \times n\_genes` if ``avg_models`` is set to
            True, or :math:`n\_hits \times n\_models \times n\_genes` if
            ``avg_models`` is set to False.
        """
        n_jobs = config.N_JOBS if n_jobs == config._USE_GLOBAL else n_jobs
        if self.query.shape[1] == 0:
            raise RuntimeError(
                "No query data available! Please set \"store_dataset\" to True "
                "when calling BLAST.query()"
            )

        ref_idx = np.concatenate(self.hits)
        query_idx = np.concatenate([
            idx * np.ones_like(_hits)
            for idx, _hits in enumerate(self.hits)
        ])
        ref = self.blast.ref[ref_idx, :]
        query = self.query[query_idx, :]

        query_latent = joblib.Parallel(
            n_jobs=min(n_jobs, len(self)), backend="threading"
        )(joblib.delayed(model.inference)(
            self.query
        ) for model in self.blast.models)
        query_latent = np.stack(query_latent)[:, query_idx, :]  # n_models * sum(n_hits) * latent_dim

        ref_latent = self.blast._get_latent(n_jobs)  # n_cells * n_models * latent_dim
        ref_latent = ref_latent[ref_idx].swapaxes(0, 1)  # n_models * sum(n_hits) * latent_dim

        deviation = query_latent - ref_latent  # n_models * sum(n_hits) * latent_dim
        if normalize_deviation:
            deviation /= np.linalg.norm(deviation, axis=2, keepdims=True)

        # TODO: We can do some sort of interpolation here to further improve
        # reliability of the results?

        if eval_point in ("ref", "both"):
            gene_dev_ref = joblib.Parallel(
                n_jobs=n_jobs, backend="threading"
            )(joblib.delayed(model.gene_grad)(
                ref, latent_grad=_deviation
            ) for model, _deviation in zip(
                self.blast.models, deviation
            ))  # n_models * [sum(n_hits) * n_genes]
            gene_dev_ref = np.stack(gene_dev_ref, axis=1)  # sum(n_hits) * n_models * n_genes
        if eval_point in ("query", "both"):
            gene_dev_query = joblib.Parallel(
                n_jobs=n_jobs, backend="threading"
            )(joblib.delayed(model.gene_grad)(
                query, latent_grad=_deviation
            ) for model, _deviation in zip(
                self.blast.models, deviation
            ))  # n_models * [sum(n_hits) * n_genes]
            gene_dev_query = np.stack(gene_dev_query, axis=1)  # sum(n_hits) * n_models * n_genes

        if eval_point == "ref":
            gene_dev = gene_dev_ref
        elif eval_point == "query":
            gene_dev = gene_dev_query
        else:  # eval_point == "both"
            gene_dev = (gene_dev_ref + gene_dev_query) / 2

        if avg_models:
            gene_dev = np.mean(gene_dev, axis=1)

        split_idx = np.cumsum([_hits.size for _hits in self.hits])[:-1]
        gene_dev = np.split(gene_dev, split_idx)  # n_queries * [n_hits * <n_models> * n_genes]
        return gene_dev


def sankey(
        query: np.ndarray, ref: np.ndarray, title: str = "Sankey",
        width: int = 500, height: int = 500, tint_cutoff: int = 1,
        font: str = "Arial", font_size: float = 10.0,
        suppress_plot: bool = False
) -> dict:  # pragma: no cover
    r"""
    Make a sankey diagram of query-reference mapping (only works in
    ipython notebooks).

    Parameters
    ----------
    query
        1-dimensional array of query annotation.
    ref
        1-dimensional array of BLAST prediction based on reference database.
    title
        Diagram title.
    width
        Graph width.
    height
        Graph height.
    tint_cutoff
        Cutoff below which sankey flows are shown in a tinter color.
    font
        Font family used for the plot.
    font_size
        Font size for the plot.
    suppress_plot
        Whether to suppress plotting and only return the figure dict.

    Returns
    -------
    fig
        Figure object fed to `iplot` of the `plotly` module to produce the plot.
    """
    cf = metrics.confusion_matrix(query, ref)
    cf["query"] = cf.index.values
    cf = cf.melt(
        id_vars=["query"], var_name="reference", value_name="count"
    ).sort_values(by="count")
    query_i, query_c = utils.encode_integer(cf["query"])
    ref_i, ref_c = utils.encode_integer(cf["reference"])

    sankey_data = dict(
        type="sankey",
        node=dict(
            pad=15,
            thickness=20,
            line=dict(
                color="black",
                width=0.5
            ),
            label=np.concatenate([
                query_c, ref_c
            ], axis=0),
            color=["#E64B35"] * len(query_c) +
            ["#4EBBD5"] * len(ref_c)
        ),
        link=dict(
            source=query_i.tolist(),
            target=(
                ref_i + len(query_c)
            ).tolist(),
            value=cf["count"].tolist(),
            color=np.vectorize(
                lambda x: "#F0F0F0" if x <= tint_cutoff else "#CCCCCC"
            )(cf["count"])
        )
    )
    sankey_layout = dict(
        title=title,
        width=width,
        height=height,
        font=dict(
            family=font,
            size=font_size
        )
    )

    fig = dict(data=[sankey_data], layout=sankey_layout)
    if not suppress_plot:
        import plotly.offline
        plotly.offline.init_notebook_mode()
        plotly.offline.iplot(fig, validate=False)
    return fig



r"""
Probabilistic / decoder modules for DIRECTi
"""

import typing
import abc

import numpy as np
import tensorflow as tf

from . import module, nn


class ProbModel(module.Module):
    r"""
    Abstract base class for generative model modules.
    """
    def __init__(
            self, h_dim: int = 128, depth: int = 1, dropout: float = 0.0,
            lambda_reg: float = 0.0, fine_tune: bool = False,
            deviation_reg: float = 0.0, name: str = "ProbModel"
    ) -> None:
        super(ProbModel, self).__init__(name=name)
        self.h_dim = h_dim
        self.depth = depth
        self.dropout = dropout
        self.lambda_reg = lambda_reg
        self.fine_tune = fine_tune
        self.deviation_reg = deviation_reg
        self.deviation_regularizer = \
            (lambda x: self.deviation_reg * tf.reduce_mean(tf.square(x))) \
            if self.fine_tune and self.deviation_reg > 0 else None

    @staticmethod
    def _normalize(  # pylint: disable=unused-argument
            x: typing.Union[np.ndarray, tf.Tensor],
            library_size: typing.Union[np.ndarray, tf.Tensor]
    ) -> typing.Union[np.ndarray, tf.Tensor]:  # pragma: no cover
        return x

    @staticmethod
    def _add_noise(  # pylint: disable=unused-argument
            x: typing.Union[np.ndarray, tf.Tensor],
            random_state: typing.Optional[np.random.RandomState] = None
    ) -> typing.Union[np.ndarray, tf.Tensor]:  # pragma: no cover
        return x

    @staticmethod
    def _preprocess(x: tf.Tensor) -> tf.Tensor:
        return x

    def _loss(
            self, ref: tf.Tensor, latent: tf.Tensor, training_flag: tf.Tensor,
            tail_concat: typing.Optional[typing.List[tf.Tensor]] = None,
            scope: str = "decoder"
    ) -> tf.Tensor:
        with tf.variable_scope(f"{scope}/{self.scope_safe_name}"):
            dropout = np.zeros(self.depth)
            dropout[1:] = self.dropout  # No dropout for first layer
            mlp_kwargs = dict(
                dropout=dropout.tolist(), dense_kwargs=dict(
                    deviation_regularizer=self.deviation_regularizer
                ), training_flag=training_flag
            )
            ptr = nn.mlp(latent, [self.h_dim] * self.depth, **mlp_kwargs)
            ptr = (ptr if isinstance(ptr, list) else [ptr]) + (tail_concat or [])
            self.log_likelihood = self._log_likelihood(ref, ptr)
            self.mean_log_likelihood = tf.reduce_mean(self.log_likelihood, axis=1)  # feature size invariant
            raw_loss = tf.negative(tf.reduce_mean(self.mean_log_likelihood), name="raw_loss")
            regularized_loss = tf.add(
                raw_loss, self.lambda_reg * self._build_regularizer(),
                name="regularized_loss"
            )
        self.vars_to_save += tf.get_collection(
            tf.GraphKeys.GLOBAL_VARIABLES, f"{scope}/{self.scope_safe_name}")
        tf.add_to_collection(tf.GraphKeys.LOSSES, raw_loss)
        tf.add_to_collection(tf.GraphKeys.LOSSES, regularized_loss)
        return regularized_loss

    @abc.abstractmethod
    def _log_likelihood(
            self, ref: tf.Tensor, pre_recon: typing.List[tf.Tensor]
    ) -> tf.Tensor:  # pragma: no cover
        raise NotImplementedError

    def _build_regularizer(self) -> tf.Tensor:
        return 0

    def _get_config(self) -> typing.Mapping:
        return {
            "h_dim": self.h_dim,
            "depth": self.depth,
            "dropout": self.dropout,
            "lambda_reg": self.lambda_reg,
            "fine_tune": self.fine_tune,
            "deviation_reg": self.deviation_reg,
            **super(ProbModel, self)._get_config()
        }

    def __bool__(self) -> bool:
        return True


class CountBased(ProbModel):  # pylint: disable=abstract-method

    @staticmethod
    def _normalize(
            x: typing.Union[np.ndarray, tf.Tensor],
            library_size: typing.Union[np.ndarray, tf.Tensor]
    ) -> typing.Union[tf.Tensor]:
        return x / (library_size / 10000)

    @staticmethod
    def _add_noise(
            x: typing.Union[np.ndarray, tf.Tensor],
            random_state: typing.Optional[np.random.RandomState] = None
    ) -> typing.Union[np.ndarray, tf.Tensor]:
        if random_state is None:
            return tf.squeeze(tf.random_poisson(x, [1]), axis=0)
        else:
            return random_state.poisson(x)

    @staticmethod
    def _preprocess(x: tf.Tensor) -> tf.Tensor:
        return tf.log1p(x)


class NB(CountBased):  # Negative binomial
    r"""
    Build a Negative Binomial generative module.

    Parameters
    ----------
    h_dim
        Dimensionality of the hidden layers in the decoder MLP.
    depth
        Number of hidden layers in the decoder MLP.
    dropout
        Dropout rate.
    fine_tune
        Whether the module is used in fine-tuning.
    lambda_reg
        Regularization strength for the generative model parameters.
        Here log-scale variance of the scale parameter
        is regularized to improve numerical stability.
    deviation_reg
        Regularization strength for the deviation from original model weights.
    name
        Name of the module.
    """
    def __init__(
            self, h_dim: int = 128, depth: int = 1, dropout: float = 0.0,
            lambda_reg: float = 0.0, fine_tune: bool = False,
            deviation_reg: float = 0.0, name: str = "NB"
    ) -> None:
        super(NB, self).__init__(
            h_dim, depth, dropout, lambda_reg,
            fine_tune, deviation_reg, name=name
        )

    def _log_likelihood(
            self, ref: tf.Tensor, pre_recon: typing.List[tf.Tensor]
    ) -> tf.Tensor:
        recon_dim = ref.get_shape().as_list()[1]
        self.softmax_mu = tf.nn.softmax(nn.dense(
            pre_recon, recon_dim,
            deviation_regularizer=self.deviation_regularizer,
            scope="softmax_mu_dense"
        ), name="softmax_mu")
        self.log_theta = tf.identity(nn.dense(
            pre_recon, recon_dim,
            deviation_regularizer=self.deviation_regularizer,
            scope="log_theta_dense"
        ), name="log_theta")
        self.recon = mu = \
            self.softmax_mu * tf.reduce_sum(ref, axis=1, keepdims=True)
        return self._log_nb_positive(ref, mu, self.log_theta)

    def _build_regularizer(self) -> tf.Tensor:
        with tf.name_scope("regularization"):
            return tf.nn.moments(self.log_theta, axes=[0, 1])[1]

    @staticmethod
    def _log_nb_positive(
            x: tf.Tensor, mu: tf.Tensor, log_theta: tf.Tensor,
            eps: float = 1e-8
    ) -> tf.Tensor:
        with tf.name_scope("log_nb_positive"):
            theta = tf.exp(log_theta)
            return theta * log_theta \
                - theta * tf.log(theta + mu + eps) \
                + x * tf.log(mu + eps) - x * tf.log(theta + mu + eps) \
                + tf.lgamma(x + theta) - tf.lgamma(theta) \
                - tf.lgamma(x + 1)


class ZINB(NB):  # Zero-inflated negative binomial
    r"""
    Build a Zero-Inflated Negative Binomial generative module.

    Parameters
    ----------
    h_dim
        Dimensionality of the hidden layers in the decoder MLP.
    depth
        Number of hidden layers in the decoder MLP.
    dropout
        Dropout rate.
    fine_tune
        Whether the module is used in fine-tuning.
    lambda_reg
        Regularization strength for the generative model parameters.
        Here log-scale variance of the scale parameter
        is regularized to improve numerical stability.
    deviation_reg
        Regularization strength for the deviation from original model weights.
    name
        Name of the module.
    """
    def __init__(
            self, h_dim: int = 128, depth: int = 1, dropout: float = 0.0,
            lambda_reg: float = 0.0, fine_tune: bool = False,
            deviation_reg: float = 0.0, name: str = "ZINB"
    ) -> None:
        super(ZINB, self).__init__(
            h_dim, depth, dropout, lambda_reg,
            fine_tune, deviation_reg, name=name
        )

    def _log_likelihood(
            self, ref: tf.Tensor, pre_recon: typing.List[tf.Tensor]
    ) -> tf.Tensor:
        recon_dim = ref.get_shape().as_list()[1]
        self.softmax_mu = tf.nn.softmax(nn.dense(
            pre_recon, recon_dim,
            deviation_regularizer=self.deviation_regularizer,
            scope="softmax_mu_dense"
        ), name="softmax_mu")
        self.log_theta = tf.identity(nn.dense(
            pre_recon, recon_dim,
            deviation_regularizer=self.deviation_regularizer,
            scope="log_theta_dense"
        ), name="log_theta")
        self.pi = tf.identity(nn.dense(
            pre_recon, recon_dim,
            deviation_regularizer=self.deviation_regularizer,
            scope="dropout_logit_dense"
        ), name="dropout_logit")
        self.recon = mu = \
            self.softmax_mu * tf.reduce_sum(ref, axis=1, keepdims=True)
        self.dropout_rate = tf.sigmoid(self.pi)
        return self._log_zinb_positive(ref, mu, self.log_theta, self.pi)

    @staticmethod
    def _log_zinb_positive(
            x: tf.Tensor, mu: tf.Tensor, log_theta: tf.Tensor,
            pi: tf.Tensor, eps: float = 1e-8
    ) -> tf.Tensor:
        r"""
        From scVI
        """
        with tf.name_scope("log_zinb_positive"):
            theta = tf.exp(log_theta)
            with tf.name_scope("case_zero"):
                case_zero = tf.nn.softplus(
                    - pi + theta * log_theta -
                    theta * tf.log(theta + mu + eps)
                ) - tf.nn.softplus(- pi)
            with tf.name_scope("case_non_zero"):
                case_non_zero = - pi - tf.nn.softplus(- pi) \
                    + theta * log_theta \
                    - theta * tf.log(theta + mu + eps) \
                    + x * tf.log(mu + eps) - x * tf.log(theta + mu + eps) \
                    + tf.lgamma(x + theta) - tf.lgamma(theta) \
                    - tf.lgamma(x + 1)
            with tf.name_scope("mixture"):
                mask = tf.cast(tf.less(x, eps), tf.float32)
                res = tf.identity(
                    tf.multiply(mask, case_zero) +
                    tf.multiply(1 - mask, case_non_zero),
                    name="likelihood")
            return res


class LN(CountBased):
    r"""
    Build a Log Normal generative module.

    Parameters
    ----------
    h_dim
        Dimensionality of the hidden layers in the decoder MLP.
    depth
        Number of hidden layers in the decoder MLP.
    dropout
        Dropout rate.
    lambda_reg
        NOT USED.
    fine_tune
        Whether the module is used in fine-tuning.
    deviation_reg
        Regularization strength for the deviation from original model weights.
    name
        Name of the module.
    """
    def __init__(
            self, h_dim: int = 128, depth: int = 1, dropout: float = 0.0,
            lambda_reg: float = 0.0, fine_tune: bool = False,
            deviation_reg: float = 0.0, name: str = "LN"
    ) -> None:
        super(LN, self).__init__(
            h_dim, depth, dropout, lambda_reg,
            fine_tune, deviation_reg, name=name
        )

    def _log_likelihood(
            self, ref: tf.Tensor, pre_recon: typing.List[tf.Tensor]
    ) -> tf.Tensor:
        recon_dim = ref.get_shape().as_list()[1]
        self.mu = tf.identity(nn.dense(
            pre_recon, recon_dim,
            deviation_regularizer=self.deviation_regularizer,
            scope="mu_dense"
        ), name="mu")
        self.log_var = tf.identity(nn.dense(
            pre_recon, recon_dim,
            deviation_regularizer=self.deviation_regularizer,
            scope="log_var_dense"
        ), name="log_var")
        self.recon = tf.expm1(self.mu)
        return self._log_ln_positive(
            tf.log1p(ref), self.mu, self.log_var)

    @staticmethod
    def _log_ln_positive(
            x: tf.Tensor, mu: tf.Tensor, log_var: tf.Tensor
    ) -> tf.Tensor:
        with tf.name_scope("log_ln"):
            return - 0.5 * (
                tf.square(x - mu) / tf.exp(log_var)
                + tf.log(2 * np.pi) + log_var
            )


class ZILN(LN):
    r"""
    Build a Zero-Inflated Log Normal generative module.

    Parameters
    ----------
    h_dim
        Dimensionality of the hidden layers in the decoder MLP.
    depth
        Number of hidden layers in the decoder MLP.
    dropout
        Dropout rate.
    lambda_reg
        NOT USED.
    fine_tune
        Whether the module is used in fine-tuning.
    deviation_reg
        Regularization strength for the deviation from original model weights.
    name
        Name of the module.
    """
    def __init__(
            self, h_dim: int = 128, depth: int = 1, dropout: float = 0.0,
            lambda_reg: float = 0.0, fine_tune: bool = False,
            deviation_reg: float = 0.0, name: str = "ZILN"
    ) -> None:
        super(ZILN, self).__init__(
            h_dim, depth, dropout, lambda_reg,
            fine_tune, deviation_reg, name=name
        )

    def _log_likelihood(
            self, ref: tf.Tensor, pre_recon: typing.List[tf.Tensor]
    ) -> tf.Tensor:
        recon_dim = ref.get_shape().as_list()[1]
        self.mu = tf.identity(nn.dense(
            pre_recon, recon_dim,
            deviation_regularizer=self.deviation_regularizer,
            scope="mu_dense"
        ), name="mu")
        self.log_var = tf.identity(nn.dense(
            pre_recon, recon_dim,
            deviation_regularizer=self.deviation_regularizer,
            scope="log_var_dense"
        ), name="log_var")
        self.pi = tf.identity(nn.dense(
            pre_recon, recon_dim,
            deviation_regularizer=self.deviation_regularizer,
            scope="dropout_logit_dense"
        ), name="dropout_logit")
        self.recon = tf.expm1(self.mu)
        self.dropout_rate = tf.sigmoid(self.pi)
        return self._log_ziln_positive(
            tf.log1p(ref), self.mu, self.log_var, self.pi)

    @staticmethod
    def _log_ziln_positive(
            x: tf.Tensor, mu: tf.Tensor, log_var: tf.Tensor,
            pi: tf.Tensor, eps: float = 1e-8
    ) -> tf.Tensor:
        with tf.name_scope("log_ziln"):
            with tf.name_scope("case_zero"):
                case_zero = - tf.nn.softplus(- pi)
            with tf.name_scope("case_non_zero"):
                case_non_zero = - pi - tf.nn.softplus(- pi) - 0.5 * (
                    tf.square(x - mu) / tf.exp(log_var)
                    + tf.log(2 * np.pi) + log_var
                )
            with tf.name_scope("mixture"):
                mask = tf.cast(tf.less(x, eps), tf.float32)
                res = tf.identity(
                    tf.multiply(mask, case_zero) +
                    tf.multiply(1 - mask, case_non_zero),
                    name="likelihood"
                )
            return res


class MSE(ProbModel):

    def __init__(
            self, h_dim: int = 128, depth: int = 1, dropout: float = 0.0,
            lambda_reg: float = 0.0, fine_tune: bool = False,
            deviation_reg: float = 0.0, name="MSE"
    ) -> None:
        super(MSE, self).__init__(
            h_dim, depth, dropout, lambda_reg,
            fine_tune, deviation_reg, name=name
        )

    def _log_likelihood(
            self, ref: tf.Tensor, pre_recon: typing.List[tf.Tensor]
    ) -> tf.Tensor:
        recon_dim = ref.get_shape().as_list()[1]
        self.mu = tf.identity(nn.dense(
            pre_recon, recon_dim,
            deviation_regularizer=self.deviation_regularizer,
            scope="mu_dense"
        ), name="mu")
        return tf.negative(tf.square(ref - self.mu))



r"""
Abstract module class for DIRECTi
"""


import os
import sys
import typing

import tensorflow as tf

from . import utils


class Module(object):

    r"""
    Build an empty module that does nothing

    Parameters
    ----------
    name
        name of the module
    """

    def __init__(self, name: typing.Optional[str] = None) -> None:
        self.name = name
        self.scope_safe_name = utils.scope_free(name)
        self.vars_to_save = []
        self.on_epoch_end = []

    def _save_weights(self, sess: tf.Session, path: str) -> None:
        if self.vars_to_save:
            if not os.path.exists(path):
                os.makedirs(path)
            tf.train.Saver(
                var_list=self.vars_to_save, max_to_keep=1
            ).save(sess, os.path.join(path, "save.ckpt"), write_meta_graph=False)

    def _load_weights(
            self, sess: tf.Session, path: str, fast: bool = False
    ) -> None:
        if fast:
            tf.train.Saver(
                var_list=self.vars_to_save, max_to_keep=1
            ).restore(sess, os.path.join(path, "save.ckpt"))
            return
        failed_vars = []
        for var_to_save in self.vars_to_save:
            try:
                tf.train.Saver(
                    var_list=[var_to_save], max_to_keep=1
                ).restore(sess, os.path.join(path, "save.ckpt"))
            except Exception:
                failed_vars.append(var_to_save)
        if failed_vars:
            utils.logger.info("%d variables failed to load.", len(failed_vars))
            utils.logger.debug(str(failed_vars))

    def _compile(self, optimizer: str, lr: float) -> None:
        pass

    @staticmethod
    def _build_feed_dict(*args, **kwargs) -> typing.Mapping:  # pylint: disable=unused-argument
        return {}

    def __bool__(self) -> bool:
        return False

    def _get_config(self) -> typing.Mapping:
        return {
            "class": ".".join((self.__module__, self.__class__.__qualname__)),
            "name": self.name
        }

    @staticmethod
    def _load_config(config: typing.Mapping) -> "Module":
        class_split = config["class"].split(".")
        module_name = ".".join(class_split[:-1])
        class_name = class_split[-1]
        c = getattr(sys.modules[module_name], class_name)
        del config["class"]
        return c(**config)



r"""
DIRECTi, an deep learning model for semi-supervised parametric dimension
reduction and systematical bias removal, extended from scVI.
"""

import json
import os
import typing
import tempfile

import numpy as np
import pandas as pd
import scipy.sparse
import tensorflow as tf

from . import config, data, latent, model, prob, rmbatch, utils

_TRAIN = 1
_TEST = 0


class DIRECTi(model.Model):
    r"""
    DIRECTi model.

    Parameters
    ----------
    genes
        Genes to use in the model.
    latent_module
        Module for latent variable (encoder module).
    prob_module
        Module for data generative modeling (decoder module).
    rmbatch_modules
        List of modules for batch effect correction.
    denoising
        Whether to add noise to the input during training (source of randomness
        in modeling the approximate posterior).
    decoder_feed_batch
        How to feed batch information to the decoder.
        Available options are listed below:
        "nonlinear": concatenate with the cell embedding vector and go through
        all nonlinear transformations in the decoder;
        "linear": concatenate with last hidden layer in the decoder and only go
        through the last linear transformation;
        "both": concatenate with both the cell embedding vector and the last
        encoder hidden layer;
        False: do not feed batch information to the decoder.
    path
        Specifies a path where model configuration, checkpoints,
        as well as the final model will be saved.
    random_seed
        Random seed. If not specified, :data:`config.RANDOM_SEED`
        will be used, which defaults to 0.

    Attributes
    ----------
    genes
        List of gene names the model is defined and fitted on

    Examples
    --------

    The :func:`fit_DIRECTi` function offers an easy to use wrapper of this
    :class:`DIRECTi` model class, which is the preferred API and should satisfy most
    needs. We suggest using the :func:`fit_DIRECTi` wrapper first.
    However, if you do wish to use this :class:`DIRECTi` class,
    here's a brief instruction:

    First you need to select the proper modules required to build a complete model.
    :class:`DIRECTi` is made up of three types of modules, a latent (encoder) module,
    a generative (decoder) module and optional batch effect correction modules:

    >>> latent_module = Cell_BLAST.latent.Gau(latent_dim=10)
    >>> prob_module = Cell_BLAST.prob.ZINB()
    >>> rmbatch_modules = [Cell_BLAST.rmbatch.Adversarial(
    ...     batch_dim=2, name="rmbatch"
    ... )]

    We also need a list of gene names which defines the gene set on which
    the model will be fitted. It can also be accessed later to help ensure
    correct gene set of the input data.
    Then we can assemble them into a complete DIRECTi model, and compile it
    (get ready for model fitting):

    >>> model = Cell_BLAST.directi.DIRECTi(
    ...     genes, latent_module, prob_module, rmbatch_modules
    ... ).compile(optimizer="RMSPropOptimizer", lr=1e-3)

    To fit the model, we need a :class:`utils.DataDict` object,
    which can be seen as a dict of array-like objects.

    In the :class:`utils.DataDict`, we require an "exprs" slot that stores the
    :math:`cell \times gene` expression matrix.
    If the latent module :class:`latent.SemiSupervisedCatGau` is used, we additionally
    require a slot containing one-hot encoded supervision labels. Unsupervised
    cells should have all zeros in the corresponding rows. Name of the slot
    should be the same as name of the semi-supervision latent module.
    If batch effect correction modules are used, we additionally require slots
    containing one-hot encoded batch information. Name of the slots should be
    the same as name of batch correction modules.

    Here's an example on how to construct a data dict from a
    :class:`data.ExprDataSet` object:

    >>> data_dict = Cell_BLAST.utils.DataDict(
    ...     exprs=data_obj[:, model.genes].exprs,
    ...     rmbatch=Cell_BLAST.utils.encode_onehot(data_obj.obs["batch"])
    ... )

    Now we are ready to fit the model:

    >>> model.fit(data_dict)

    At this point we have obtained a fitted DIRECTi model object (which is also
    the return value if the :func:`fit_DIRECTi` function).
    We can use this model to project transriptomes into the low dimensional
    cell embedding space by using the :meth:`DIRECTi.inference` method.
    You may pass it to :attr:`data.ExprDataSet.latent` slot in the original
    data object, which facilitates subsequent visualization of the latent
    embedding space:

    >>> data_obj.latent = model.inference(data_obj)
    >>> data_obj.visualize_latent("cell_type")
    """

    _TRAIN = 1
    _TEST = 0

    def __init__(
            self, genes: typing.List[str],
            latent_module: latent.Latent,
            prob_module: prob.ProbModel,
            rmbatch_modules: typing.Optional[typing.List[rmbatch.RMBatch]] = None,
            denoising: bool = True,
            decoder_feed_batch: typing.Union[str, bool] = "nonlinear",
            path: typing.Optional[str] = None,
            random_seed: int = config._USE_GLOBAL,
            _mode: int = _TRAIN
    ) -> None:
        random_seed = config.RANDOM_SEED \
            if random_seed == config._USE_GLOBAL else random_seed
        if isinstance(genes, pd.Series):
            genes = genes.values
        if isinstance(genes, np.ndarray):
            genes = genes.tolist()
        assert isinstance(genes, list)
        self.genes = genes
        self.x_dim = len(genes)
        self._mode = _mode
        super(DIRECTi, self).__init__(
            latent_module=latent_module,
            prob_module=prob_module,
            rmbatch_modules=rmbatch_modules,
            denoising=denoising,
            decoder_feed_batch=decoder_feed_batch,
            random_seed=random_seed,
            path=path
        )

    def _init_graph(
            self, latent_module: latent.Latent, prob_module: prob.ProbModel,
            rmbatch_modules: typing.Optional[typing.List[rmbatch.RMBatch]] = None,
            denoising: bool = True,
            decoder_feed_batch: typing.Union[str, bool] = "nonlinear"
    ) -> None:
        super(DIRECTi, self)._init_graph()
        self.denoising = denoising
        self.decoder_feed_batch = decoder_feed_batch
        self.latent_module = latent_module
        self.prob_module = prob_module
        self.rmbatch_modules = rmbatch_modules or []
        self.loss, self.early_stop_loss = [], []

        with tf.name_scope("placeholder/"):
            self.x = tf.placeholder(
                dtype=tf.float32, shape=(None, self.x_dim), name="x")
            self.library_size = tf.placeholder(
                dtype=tf.float32, shape=(None, 1), name="library_size")
            self.training_flag = tf.placeholder(
                dtype=tf.bool, shape=(), name="training_flag")

        # Preprocessing
        with tf.name_scope("normalize"):
            normalized_x = prob_module._normalize(self.x, self.library_size)
        with tf.name_scope("noise"):
            self.noisy_x = prob_module._add_noise(normalized_x) \
                if denoising else normalized_x
        with tf.name_scope("preprocess"):
            self.preprocessed_x = prob_module._preprocess(self.noisy_x)

        # Encoder
        self.latent = self.latent_module._build_latent(
            self.preprocessed_x, self.training_flag, scope="encoder")
        if self._mode == _TEST:
            return
        self.loss.append(self.latent_module._build_regularizer(
            self.training_flag, self.epoch))

        # Remove batch effect
        for rmbatch_module in self.rmbatch_modules:
            self.loss.append(rmbatch_module._build_regularizer(
                self.latent, self.training_flag, self.epoch))

        # Decoder
        feed_batch = [
            rmbatch_module.batch for rmbatch_module in self.rmbatch_modules
        ] if decoder_feed_batch in (
            "nonlinear", "linear", "both"
        ) and self.rmbatch_modules else None
        full_latent = [self.latent] + feed_batch if decoder_feed_batch in (
            "nonlinear", "both"
        ) and self.rmbatch_modules else [self.latent]
        tail_concat = feed_batch if decoder_feed_batch in (
            "linear", "both"
        ) and self.rmbatch_modules else None
        self.recon_loss = self.prob_module._loss(
            self.x, full_latent, self.training_flag,
            tail_concat=tail_concat, scope="decoder"
        )

        self.loss.append(self.recon_loss)
        self.early_stop_loss.append(self.recon_loss)

        self.loss = tf.add_n([
            item for item in
            self.loss + tf.get_collection(tf.GraphKeys.REGULARIZATION_LOSSES)
            if item != 0
        ], name="total_loss")
        self.early_stop_loss = tf.add_n([
            item for item in self.early_stop_loss if item != 0
        ], name="early_stop_loss")
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.early_stop_loss)
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.loss)

        self.grad_dict = {}

    def _compile(
            self, optimizer: str = "RMSPropOptimizer", lr: float = 1e-3
    ) -> None:
        if self.latent_module:
            self.latent_module._compile(optimizer, lr)
        if self.prob_module:
            self.prob_module._compile(optimizer, lr)
        for rmbatch_module in self.rmbatch_modules:
            rmbatch_module._compile(optimizer, lr)
        with tf.variable_scope("optimize/main"):
            control_dependencies = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
            with tf.control_dependencies(control_dependencies):
                self.step = getattr(tf.train, optimizer)(lr).minimize(
                    self.loss, var_list=tf.get_collection(
                        tf.GraphKeys.TRAINABLE_VARIABLES, "encoder"
                    ) + tf.get_collection(
                        tf.GraphKeys.TRAINABLE_VARIABLES, "decoder"
                    )
                )

    def _fit_epoch(
            self, data_dict: utils.DataDict, batch_size: int = 128,
            progress_bar: bool = True
    ) -> None:
        loss_dict = {
            item: 0.0 for item in tf.get_collection(tf.GraphKeys.LOSSES)
        }

        # Training
        @utils.minibatch(batch_size, desc="training",
                         use_last=False, progress_bar=progress_bar)
        def _train(data_dict):
            nonlocal loss_dict
            feed_dict = {
                self.x: utils.densify(data_dict["exprs"]),
                self.library_size: data_dict["library_size"],
                **(self.latent_module._build_feed_dict(data_dict)),
                **(self.prob_module._build_feed_dict(data_dict)),
                self.training_flag: True
            }
            for rmbatch_module in self.rmbatch_modules:
                feed_dict.update(rmbatch_module._build_feed_dict(data_dict))
            run_result = self.sess.run(
                [self.step] + list(loss_dict.keys()), feed_dict=feed_dict)
            run_result.pop(0)
            for item in loss_dict:
                loss_dict[item] += run_result.pop(0) * data_dict.size

        _train(data_dict)
        for item in loss_dict:
            loss_dict[item] /= data_dict.size
        self.epoch_report += f"train={loss_dict[self.early_stop_loss]:.3f}, "

        manual_summary = tf.Summary(value=[
            tf.Summary.Value(tag=f"{item.name} (train)",
                             simple_value=loss_dict[item])
            for item in loss_dict
        ])
        self.summarizer.add_summary(manual_summary, self.sess.run(self.epoch))

    def _val_epoch(
            self, data_dict: utils.DataDict, batch_size: int = 128,
            progress_bar: bool = True
    ) -> float:
        loss_dict = {
            item: 0.0 for item in tf.get_collection(tf.GraphKeys.LOSSES)
        }

        # Validation
        @utils.minibatch(batch_size, desc="validation",
                         use_last=True, progress_bar=progress_bar)
        def _validate(data_dict):
            nonlocal loss_dict
            feed_dict = {
                self.x: utils.densify(data_dict["exprs"]),
                self.library_size: data_dict["library_size"],
                **(self.latent_module._build_feed_dict(data_dict)),
                **(self.prob_module._build_feed_dict(data_dict)),
                self.training_flag: False
            }
            for rmbatch_module in self.rmbatch_modules:
                feed_dict.update(rmbatch_module._build_feed_dict(data_dict))
            run_result = self.sess.run(
                list(loss_dict.keys()),
                feed_dict=feed_dict)
            for item in loss_dict:
                loss_dict[item] += run_result.pop(0) * data_dict.size

        _validate(data_dict)
        for item in loss_dict:
            loss_dict[item] /= data_dict.size
        self.epoch_report += f"val={loss_dict[self.early_stop_loss]:.3f}, "

        manual_summary = tf.Summary(value=[
            tf.Summary.Value(tag=f"{item.name} (val)",
                             simple_value=loss_dict[item])
            for item in loss_dict
        ])
        self.summarizer.add_summary(manual_summary, self.sess.run(self.epoch))
        return loss_dict[self.early_stop_loss]

    def fit(
            self, data_dict: utils.DataDict, batch_size: int = 128,
            val_split: float = 0.1, epoch: int = 1000,
            patience: int = 30, tolerance: float = 0.0,
            on_epoch_end: typing.Optional[typing.List[typing.Callable[
                ["DIRECTi", utils.DataDict, utils.DataDict, float], bool
            ]]] = None,
            progress_bar: bool = False
    ) -> "DIRECTi":
        r"""
        Fit the model.

        Parameters
        ----------
        data_dict
            Training data.
        batch_size
            Size of minibatch used in training.
        val_split
            Fraction of data to use for validation.
        epoch
            Maximal training epochs.
        patience
            Early stop patience. Model training stops when the
            best validation loss does not decrease for a consecutive
            ``patience`` epochs.
        on_epoch_end
            List of functions to be executed at the end of each epoch.
        progress_bar
            Whether to print progress bar for each epoch during training.

        Returns
        -------
        model
            The fitted model
        """
        on_epoch_end = on_epoch_end or []
        on_epoch_end += \
            self.latent_module.on_epoch_end + self.prob_module.on_epoch_end
        for rmbatch_module in self.rmbatch_modules:
            on_epoch_end += rmbatch_module.on_epoch_end
        return super(DIRECTi, self).fit(
            data_dict, batch_size=batch_size, val_split=val_split, epoch=epoch,
            patience=patience, tolerance=tolerance, on_epoch_end=on_epoch_end,
            progress_bar=progress_bar
        )

    @utils.with_self_graph
    def _fetch(
            self, tensor: tf.Tensor,
            data_dict: typing.Optional[utils.DataDict] = None,
            batch_size: int = 4096, noisy: bool = False,
            progress_bar: bool = False, random_seed: int = config._USE_GLOBAL
    ) -> np.ndarray:
        if data_dict is None:
            return self.sess.run(tensor)
        if noisy:
            random_seed = config.RANDOM_SEED \
                if random_seed == config._USE_GLOBAL else random_seed
            random_state = np.random.RandomState(seed=random_seed)
        result_shape = tensor.get_shape().as_list()
        if result_shape[0] is None:
            result_shape[0] = data_dict.shape[0]
        result = np.empty(result_shape)

        @utils.minibatch(batch_size, desc="fetch", use_last=True,
                         progress_bar=progress_bar)
        def _fetch_minibatch(data_dict, result):
            feed_dict = {self.training_flag: False}
            if "exprs" in data_dict and "library_size" in data_dict:
                x = data_dict["exprs"]
                normalized_x = self.prob_module._normalize(x, data_dict["library_size"])
                feed_dict.update({
                    self.x: utils.densify(x),
                    self.noisy_x: self.prob_module._add_noise(
                        utils.densify(normalized_x), random_state
                    ) if noisy else utils.densify(normalized_x),
                    self.library_size: data_dict["library_size"]
                })
                # Tensorflow random samplers are fixed after creation,
                # making it impossible to re-seed and generate reproducible
                # results, so we use numpy samplers instead.
                # Also, local RandomState object is used to ensure thread-safety.
            for module in [self.latent_module, self.prob_module, *self.rmbatch_modules]:
                try:
                    feed_dict.update(module._build_feed_dict(data_dict))
                except Exception:
                    pass
            result[:] = self.sess.run(tensor, feed_dict=feed_dict)

        _fetch_minibatch(data_dict, result)
        return result

    def inference(
            self, dataset: data.ExprDataSet, batch_size: int = 4096,
            n_posterior: int = 0, progress_bar: bool = False,
            priority: str = "auto", random_seed: int = config._USE_GLOBAL
    ) -> np.ndarray:
        r"""
        Project expression profiles into the cell embedding space.

        Parameters
        ----------
        x
            Dataset for which to compute cell embeddings.
        batch_size
            Minibatch size.
            Changing this may slighly affect speed, but not the result.
        n_posterior
            How many posterior samples to fetch.
            If set to 0, the posterior point estimate is computed.
            If greater than 0, produces ``n_posterior`` number of
            posterior samples for each cell.
        progress_bar
            Whether to show progress bar duing projection.
        priority
            Should be among {"auto", "speed", "memory"}.
            Controls which one of speed or memory should be prioritized, by
            default "auto", meaning that data with more than 100,000 cells will
            use "memory" mode and smaller data will use "speed" mode.
        random_seed
            Random seed used with noisy projection. If not specified,
            :data:`config.RANDOM_SEED` will be used, which defaults to 0.

        Returns
        -------
        latent
            Coordinates in the latent space.
            If ``n_posterior`` is 0, will be in shape :math:`cell \times latent\_dim`.
            If ``n_posterior`` is greater than 0, will be in shape
            :math:`cell \times noisy \times latent\_dim`.
        """
        random_seed = config.RANDOM_SEED \
            if random_seed == config._USE_GLOBAL else random_seed
        x = dataset[:, self.genes].exprs
        l = np.array(dataset.exprs.sum(axis=1)).reshape((-1, 1)) \
            if "__libsize__" not in dataset.obs.columns \
            else dataset.obs["__libsize__"].values.reshape((-1, 1))
        if n_posterior > 0:
            if priority == "auto":
                priority = "memory" if x.shape[0] > 1e4 else "speed"
            if priority == "speed":
                if scipy.sparse.issparse(x):
                    xrep = x.tocsr()[np.repeat(np.arange(x.shape[0]), n_posterior)]
                else:
                    xrep = np.repeat(x, n_posterior, axis=0)
                lrep = np.repeat(l, n_posterior, axis=0)
                data_dict = utils.DataDict(exprs=xrep, library_size=lrep)
                return self._fetch(
                    self.latent, data_dict, batch_size, True, progress_bar, random_seed
                ).astype(np.float32).reshape((x.shape[0], n_posterior, -1))
            else:  # priority == "memory":
                data_dict = utils.DataDict(exprs=x, library_size=l)
                return np.stack([self._fetch(
                    self.latent, data_dict, batch_size, True, progress_bar,
                    (random_seed + i) if random_seed is not None else None
                ).astype(np.float32) for i in range(n_posterior)], axis=1)
        data_dict = utils.DataDict(exprs=x, library_size=l)
        return self._fetch(
            self.latent, data_dict, batch_size, False, progress_bar
        ).astype(np.float32)

    def clustering(
            self, dataset: data.ExprDataSet, batch_size: int = 4096,
            return_confidence: bool = False, progress_bar: bool = False
    ) -> typing.Tuple[np.ndarray, np.ndarray]:
        r"""
        Get model intrinsic clustering of the data.

        Parameters
        ----------
        x
            Dataset for which to obtain the intrinsic clustering.
        batch_size
            Minibatch size.
            Changing this may slighly affect speed, but not the result.
        progress_bar
            Whether to show progress bar during projection.

        Returns
        -------
        idx
            model intrinsic clustering index, 1 dimensional
        confidence
            model intrinsic clustering confidence, 1 dimensional
        """
        if not isinstance(self.latent_module, latent.CatGau):
            raise Exception("Model has no intrinsic clustering")
        x = dataset[:, self.genes].exprs
        l = np.array(dataset.exprs.sum(axis=1)).reshape((-1, 1)) \
            if "__libsize__" not in dataset.obs.columns \
            else dataset.obs["__libsize__"].values.reshape((-1, 1))
        data_dict = utils.DataDict(exprs=x, library_size=l)
        cat = self._fetch(
            self.latent_module.cat, data_dict, batch_size, False, progress_bar
        ).astype(np.float32)
        if return_confidence:
            return cat.argmax(axis=1), cat.max(axis=1)
        return cat.argmax(axis=1)

    @utils.with_self_graph
    def _fetch_grad(
            self, input_tensor: tf.Tensor, output_tensor: tf.Tensor,
            data_dict: utils.DataDict, batch_size: int = 4096,
            progress_bar: bool = False
    ) -> np.ndarray:
        r"""
        Requires "output_grad" slot in data_dict as the gradient source.
        Additionally, it requires either explicit value of the output tensor (as
        "output" slot in data_dict), or sufficient data to compute the output
        tensor (e.g., "exprs" and "library_size" slots if output tensor
        is the latent variable).
        """
        input_scope_safe_name = utils.scope_free(input_tensor.name)
        output_scope_safe_name = utils.scope_free(output_tensor.name)
        with tf.name_scope("custom_grad/"):
            if output_tensor not in self.grad_dict:
                self.grad_dict[output_tensor] = tf.placeholder(
                    dtype=tf.float32,
                    shape=output_tensor.shape,
                    name=f"{output_scope_safe_name}_grad"
                )
            if (input_tensor, output_tensor) not in self.grad_dict:
                self.grad_dict[(input_tensor, output_tensor)] = tf.gradients(
                    output_tensor, input_tensor,
                    grad_ys=self.grad_dict[output_tensor],
                    name=f"{input_scope_safe_name}_grad_from_{output_scope_safe_name}"
                )[0]

        result_shape = input_tensor.get_shape().as_list()
        if result_shape[0] is None:
            result_shape[0] = data_dict.shape[0]
        result = np.empty(result_shape)

        @utils.minibatch(batch_size, desc="fetch_grad", use_last=True,
                         progress_bar=progress_bar)
        def _fetch_grad_minibatch(data_dict, result):
            feed_dict = {
                self.grad_dict[output_tensor]: data_dict["output_grad"],
                self.training_flag: False
            }
            if "output" in data_dict:
                feed_dict[output_tensor] = data_dict["output"]
            if "exprs" in data_dict and "library_size" in data_dict:
                x = data_dict["exprs"]
                normalized_x = self.prob_module._normalize(x, data_dict["library_size"])
                feed_dict.update({
                    self.x: utils.densify(x),
                    self.noisy_x: utils.densify(normalized_x),
                    self.library_size: data_dict["library_size"]
                })
            for module in [self.latent_module, self.prob_module, *self.rmbatch_modules]:
                try:
                    feed_dict.update(module._build_feed_dict(data_dict))
                except Exception:
                    pass
            result[:] = self.sess.run(
                self.grad_dict[(input_tensor, output_tensor)],
                feed_dict=feed_dict
            )

        _fetch_grad_minibatch(data_dict, result)
        return result

    @utils.with_self_graph
    def gene_grad(
            self, dataset: data.ExprDataSet, latent_grad: np.ndarray,
            batch_size: int = 4096, progress_bar: bool = False
    ) -> np.ndarray:
        r"""
        Fetch gene space gradients with regard to latent space gradients

        Parameters
        ----------
        dataset
            Dataset for which to obtain gene gradients.
        latent_grad
            Latent space gradients.
        batch_size
            Minibatch size.
            Changing this may slighly affect speed, but not the result.
        progress_bar
            Whether to show progress bar during projection.

        Returns
        -------
        grad
            Fetched gene-wise gradient
        """
        x = dataset[:, self.genes].exprs
        l = np.array(dataset.exprs.sum(axis=1)).reshape((-1, 1)) \
            if "__libsize__" not in dataset.obs.columns \
            else dataset.obs["__libsize__"].values.reshape((-1, 1))
        data_dict = utils.DataDict(
            exprs=x, library_size=l, output_grad=latent_grad)
        return self._fetch_grad(
            self.preprocessed_x, self.latent, data_dict,
            batch_size=batch_size, progress_bar=progress_bar
        )

    def _save_weights(self, path: str) -> None:
        super(DIRECTi, self)._save_weights(os.path.join(path, "main"))
        with self.graph.as_default():  # pylint: disable=not-context-manager
            self.latent_module._save_weights(
                self.sess, os.path.join(path, "latent"))
            self.prob_module._save_weights(
                self.sess, os.path.join(path, "prob"))
            for rmbatch_module in self.rmbatch_modules:
                rmbatch_module._save_weights(self.sess, os.path.join(
                    path, "rmbatch", rmbatch_module.name))

    def _load_weights(self, path: str) -> None:
        super(DIRECTi, self)._load_weights(os.path.join(path, "main"))
        with self.graph.as_default():  # pylint: disable=not-context-manager
            utils.logger.info("Loading latent module weights...")
            self.latent_module._load_weights(self.sess, os.path.join(
                path, "latent"
            ), fast=self._mode == _TEST)
            if self._mode == _TEST:
                return
            utils.logger.info("Loading prob module weights...")
            self.prob_module._load_weights(self.sess, os.path.join(path, "prob"))
            utils.logger.info("Loading rmbatch module weights...")
            for rmbatch_module in self.rmbatch_modules:
                rmbatch_module._load_weights(self.sess, os.path.join(
                    path, "rmbatch", rmbatch_module.name
                ))
            self.sess.run(tf.get_collection(tf.GraphKeys.READY_OP))

    def _get_config(self) -> typing.Mapping:
        config = {
            "genes": self.genes,
            "denoising": self.denoising,
            "decoder_feed_batch": self.decoder_feed_batch,
            "latent_module": self.latent_module._get_config(),
            "prob_module": self.prob_module._get_config(),
            "rmbatch_modules": [],
            **super(DIRECTi, self)._get_config()
        }
        for rmbatch_module in self.rmbatch_modules:
            config["rmbatch_modules"].append(rmbatch_module._get_config())
        return config

    @classmethod
    def _load_config(cls, file, **kwargs) -> "DIRECTi":
        with open(file, "r") as f:
            config = json.load(f)
        rmbatch_modules = []
        if "rmbatch_modules" in config:
            for rmbatch_module in config["rmbatch_modules"]:
                rmbatch_modules.append(
                    rmbatch.RMBatch._load_config(rmbatch_module)
                )
        model = cls(
            genes=config["genes"],
            denoising=config["denoising"],
            decoder_feed_batch=config["decoder_feed_batch"],
            latent_module=latent.Latent._load_config(config["latent_module"]),
            prob_module=prob.ProbModel._load_config(config["prob_module"]),
            rmbatch_modules=rmbatch_modules,
            path=os.path.dirname(file), **kwargs
        )
        with model.graph.as_default():
            model.sess.run(tf.global_variables_initializer())
        return model


def fit_DIRECTi(
        dataset: data.ExprDataSet,
        genes: typing.Optional[typing.List[str]] = None,
        supervision: typing.Optional[str] = None,
        batch_effect: typing.Optional[typing.List[str]] = None,
        latent_dim: int = 10, cat_dim: typing.Optional[int] = None,
        h_dim: int = 128, depth: int = 1, prob_module: str = "NB",
        rmbatch_module: typing.Union[str, typing.List[str]] = "Adversarial",
        latent_module_kwargs: typing.Optional[typing.Mapping] = None,
        prob_module_kwargs: typing.Optional[typing.Mapping] = None,
        rmbatch_module_kwargs: typing.Optional[typing.Union[
            typing.Mapping, typing.List[typing.Mapping]
        ]] = None,
        optimizer: str = "RMSPropOptimizer", learning_rate: float = 1e-3,
        batch_size: int = 128, val_split: float = 0.1, epoch: int = 1000,
        patience: int = 30, progress_bar: bool = False, reuse_weights=None,
        random_seed: int = config._USE_GLOBAL, path: typing.Optional[str] = None
) -> DIRECTi:
    r"""
    A convenient one-step function to build and fit DIRECTi models.
    Should work well in most cases.

    Parameters
    ----------
    dataset
        Dataset to be fitted.
    genes
        Genes to fit on, should be a subset of :attr:`data.ExprDataSet.var_names`.
        If not specified, all genes are used.
    supervision
        Specifies a column in the :attr:`data.ExprDataSet.obs` table for use as
        (semi-)supervision. If value in the specified column is emtpy,
        the corresponding cells will be treated as unsupervised.
    batch_effect
        Specifies one or more columns in the :attr:`data.ExprDataSet.obs` table
        for use as batch effect to be corrected.
    latent_dim
        Latent space (cell embedding) dimensionality.
    cat_dim
        Number of intrinsic clusters.
    h_dim
        Hidden layer dimensionality. It is used consistently across all MLPs
        in the model.
    depth
        Hidden layer depth. It is used consistently across all MLPs in the model.
    prob_module
        Generative model to fit, should be among {"NB", "ZINB", "LN", "ZILN"}.
        See the :mod:`prob` for details.
    rmbatch_module
        Batch effect correction method. If a list is provided, each element
        specifies the method to use for a corresponding batch effect in
        ``batch_effect`` list (in this case the ``rmbatch_module`` list should
        have the same length as the ``batch_effect`` list).
    latent_module_kwargs
        Keyword arguments to be passed to the latent module.
    prob_module_kwargs
        Keyword arguments to be passed to the prob module.
    rmbatch_module_kwargs
        Keyword arguments to be passed to the rmbatch module.
        If a list is provided, each element specifies keyword arguments
        for a corresponding batch effect correction module in the
        ``rmbatch_module`` list.
    optimizer
        Name of optimizer used in training.
    learning_rate
        Learning rate used in training.
    batch_size
        Size of minibatch used in training.
    val_split
        Fraction of data to use for validation.
    epoch
        Maximal training epochs.
    patience
        Early stop patience. Model training stops when best validation loss does
        not decrease for a consecutive ``patience`` epochs.
    progress_bar
        Whether to show progress bars during training.
    reuse_weights
        Specifies a path where previously stored model weights can be reused.
    random_seed
        Random seed. If not specified, :data:`config.RANDOM_SEED`
        will be used, which defaults to 0.
    path
        Specifies a path where model checkpoints as well as the final model
        will be saved.

    Returns
    -------
    model
        A fitted DIRECTi model.

    Examples
    --------
    See the DIRECTi ipython notebook (:ref:`vignettes`) for live examples.
    """
    random_seed = config.RANDOM_SEED \
        if random_seed == config._USE_GLOBAL else random_seed
    if latent_module_kwargs is None:
        latent_module_kwargs = {}
    if prob_module_kwargs is None:
        prob_module_kwargs = {}
    if rmbatch_module_kwargs is None:
        rmbatch_module_kwargs = {}

    if genes is None:
        genes = dataset.var_names.values
    data_dict = utils.DataDict(
        library_size=np.array(dataset.exprs.sum(axis=1)).reshape((-1, 1))
        if "__libsize__" not in dataset.obs.columns
        else dataset.obs["__libsize__"].values.reshape((-1, 1)),
        exprs=dataset[:, genes].exprs
    )

    if batch_effect is None:
        batch_effect = []
    elif isinstance(batch_effect, str):
        batch_effect = [batch_effect]
    for _batch_effect in batch_effect:
        data_dict[_batch_effect] = utils.encode_onehot(
            dataset.obs[_batch_effect].astype(object).fillna("IgNoRe"),
            sort=True, ignore="IgNoRe"
        )  # sorting ensures batch order reproducibility for later tuning

    if supervision is not None:
        data_dict[supervision] = utils.encode_onehot(
            dataset.obs[supervision].astype(object).fillna("IgNoRe"),
            sort=True, ignore="IgNoRe"
        )  # sorting ensures supervision order reproducibility for later tuning
        if cat_dim is None:
            cat_dim = data_dict[supervision].shape[1]
        elif cat_dim > data_dict[supervision].shape[1]:
            data_dict[supervision] = scipy.sparse.hstack([
                data_dict[supervision].tocsc(),
                scipy.sparse.csc_matrix((
                    data_dict[supervision].shape[0],
                    cat_dim - data_dict[supervision].shape[1]
                ))
            ]).tocsr()
        elif cat_dim < data_dict[supervision].shape[1]:  # pragma: no cover
            raise ValueError(
                "`cat_dim` must be greater than or equal to "
                "number of supervised classes!"
            )
        # else ==

    kwargs = dict(latent_dim=latent_dim, h_dim=h_dim, depth=depth)
    if cat_dim:
        kwargs.update(dict(cat_dim=cat_dim))
        if supervision:
            kwargs.update(dict(name=supervision))
            kwargs.update(latent_module_kwargs)
            latent_module = latent.SemiSupervisedCatGau(**kwargs)
        else:
            kwargs.update(latent_module_kwargs)
            latent_module = latent.CatGau(**kwargs)
    else:
        kwargs.update(latent_module_kwargs)
        latent_module = latent.Gau(**kwargs)

    if not isinstance(rmbatch_module, list):
        rmbatch_module = [rmbatch_module] * len(batch_effect)
    if not isinstance(rmbatch_module_kwargs, list):
        rmbatch_module_kwargs = [rmbatch_module_kwargs] * len(batch_effect)
    assert len(rmbatch_module_kwargs) == len(rmbatch_module) == len(batch_effect)

    rmbatch_list = []
    for _batch_effect, _rmbatch_module, _rmbatch_module_kwargs in zip(
            batch_effect, rmbatch_module, rmbatch_module_kwargs
    ):
        kwargs = dict(
            batch_dim=data_dict[_batch_effect].shape[1],
            name=_batch_effect
        )
        if _rmbatch_module in (
                "Adversarial", "MNNAdversarial", "AdaptiveMNNAdversarial"
        ):
            kwargs.update(dict(h_dim=h_dim, depth=depth))
            kwargs.update(_rmbatch_module_kwargs)
        elif _rmbatch_module not in ("RMBatch", "MNN"):  # pragma: no cover
            raise ValueError("Invalid rmbatch method!")
        # else "RMBatch" or "MNN"
        kwargs.update(_rmbatch_module_kwargs)
        rmbatch_list.append(getattr(rmbatch, _rmbatch_module)(**kwargs))

    kwargs = dict(h_dim=h_dim, depth=depth)
    kwargs.update(prob_module_kwargs)
    prob_module = getattr(prob, prob_module)(**kwargs)
    model = DIRECTi(
        genes=genes,
        latent_module=latent_module,
        prob_module=prob_module,
        rmbatch_modules=rmbatch_list,
        path=path,
        random_seed=random_seed
    ).compile(optimizer=optimizer, lr=learning_rate)
    if reuse_weights is not None:
        model._load_weights(reuse_weights)
    model.fit(
        data_dict, batch_size=batch_size, val_split=val_split,
        epoch=epoch, patience=patience, progress_bar=progress_bar
    )
    return model


def align_DIRECTi(
        model: DIRECTi, original_dataset: data.ExprDataSet,
        new_dataset: typing.Union[data.ExprDataSet, typing.Mapping[str, data.ExprDataSet]],
        rmbatch_module: str = "MNNAdversarial",
        rmbatch_module_kwargs: typing.Optional[typing.Mapping] = None,
        deviation_reg: float = 0.01, batch_size: int = 256, val_split: float = 0.1,
        optimizer: str = "RMSPropOptimizer", learning_rate: float = 1e-3,
        epoch: int = 100, patience: int = 100, tolerance: float = 0.0,
        reuse_weights: bool = True, progress_bar: bool = False,
        random_seed: int = config._USE_GLOBAL, path: typing.Optional[str] = None
) -> DIRECTi:
    r"""
    Align datasets starting with an existing DIRECTi model (fine-tuning)

    Parameters
    ----------
    model
        A pretrained DIRECTi model.
    original_dataset
        The dataset that the model was originally trained on.
    new_dataset
        A new dataset or a dictionary containing new datasets,
        to be aligned with ``original_dataset``.
    rmbatch_module
        Specifies the batch effect correction method to use for aligning new
        datasets.
    rmbatch_module_kwargs
        Keyword arguments to be passed to the rmbatch module.
    deviation_reg
        Regularization strength for the deviation from original model weights.
    batch_size
        Size of minibatches used in training.
    val_split
        Fraction of data to use for validation.
    optimizer
        Name of the optimizer to use.
    learning_rate
        Learning rate.
    epoch
        Maximal training epochs.
    patience
        Early stop patience. Model training stops when best
        validation loss does not decrease for a consecutive ``patience`` epochs.
    tolerance
        Tolerance of deviation from the lowest validation loss recorded for the
        "patience countdown" to be reset. The "patience countdown" is reset if
        current validation loss < lowest validation loss recorded + ``tolerance``.
    reuse_weights
        Whether to reuse weights of the original model.
    progress_bar
        Whether to show progress bar during training.
    random_seed
        Random seed. If not specified, :data:`config.RANDOM_SEED`
        will be used, which defaults to 0.
    path
        Specifies a path where model checkpoints as well as the final model
        is saved.

    Returns
    -------
    aligned_model
        Aligned model.
    """
    random_seed = config.RANDOM_SEED \
        if random_seed == config._USE_GLOBAL else random_seed
    if path is None:
        path = tempfile.mkdtemp()
    else:
        os.makedirs(path, exist_ok=True)
    if rmbatch_module_kwargs is None:
        rmbatch_module_kwargs = {}
    if isinstance(new_dataset, data.ExprDataSet):
        new_datasets = {"__new__": new_dataset}
    elif isinstance(new_dataset, dict):
        assert "__original__" not in new_dataset, \
            "Key `__original__` is now allowed in new datasets."
        new_datasets = new_dataset.copy()  # shallow
    else:
        raise TypeError("Invalid type for argument `new_dataset`.")

    _config = model._get_config()
    for _rmbatch_module in _config["rmbatch_modules"]:
        _rmbatch_module["delay"] = 0
    kwargs = {
        "class": f"Cell_BLAST.rmbatch.{rmbatch_module}",
        "batch_dim": len(new_datasets) + 1,
        "delay": 0, "name": "__align__"
    }
    if rmbatch_module in (
            "Adversarial", "MNNAdversarial", "AdaptiveMNNAdversarial"
    ):
        kwargs.update(dict(
            h_dim=model.latent_module.h_dim,
            depth=model.latent_module.depth,
            dropout=model.latent_module.dropout,
            lambda_reg=0.01
        ))
    elif rmbatch_module not in ("RMBatch", "MNN"):  # pragma: no cover
        raise ValueError("Unknown rmbatch_module!")
    # else "RMBatch" or "MNN"
    kwargs.update(rmbatch_module_kwargs)
    _config["rmbatch_modules"].append(kwargs)

    _config["prob_module"]["fine_tune"] = True
    _config["prob_module"]["deviation_reg"] = deviation_reg

    with open(os.path.join(path, "config.json"), "w") as f:
        json.dump(_config, f, indent=4)
    aligned_model = DIRECTi._load_config(
        os.path.join(path, "config.json"), random_seed=random_seed
    ).compile(optimizer, learning_rate)
    if reuse_weights:
        model._save_weights(os.path.join(path, "unaligned"))
        aligned_model._load_weights(os.path.join(path, "unaligned"))
    supervision = aligned_model.latent_module.name if isinstance(
        aligned_model.latent_module, latent.SemiSupervisedCatGau
    ) else None

    assert "__align__" not in original_dataset.obs.columns, \
        "Please remove column `__align__` from obs of the original dataset."
    original_dataset = original_dataset.copy()  # shallow
    original_dataset.obs = original_dataset.obs.copy(deep=False)
    if "__libsize__" not in original_dataset.obs.columns:
        original_dataset.obs["__libsize__"] = np.array(original_dataset.exprs.sum(axis=1)).ravel()
    original_dataset = original_dataset[:, model.genes]
    for key in new_datasets.keys():
        assert "__align__" not in new_datasets[key].obs.columns, \
            f"Please remove column `__align__` from new dataset {key}."
        new_datasets[key] = new_datasets[key].copy()  # shallow
        new_datasets[key].obs = new_datasets[key].obs.copy(deep=False)
        new_datasets[key].obs = new_datasets[key].obs.loc[:, new_datasets[key].obs.columns == "__libsize__"]
        if "__libsize__" not in new_datasets[key].obs.columns:
            new_datasets[key].obs["__libsize__"] = np.array(new_datasets[key].exprs.sum(axis=1)).ravel()
        new_datasets[key] = new_datasets[key][:, model.genes]
        # All meta in new datasets are cleared to avoid interference
    dataset = {"__original__": original_dataset, **new_datasets}
    dataset = data.ExprDataSet.merge_datasets(dataset, meta_col="__align__")

    data_dict = utils.DataDict(
        library_size=dataset.obs["__libsize__"].values.reshape((-1, 1)),
        exprs=dataset[:, model.genes].exprs  # Ensure order
    )
    for rmbatch_module in aligned_model.rmbatch_modules:
        data_dict[rmbatch_module.name] = utils.encode_onehot(
            dataset.obs[rmbatch_module.name].astype(object).fillna("IgNoRe"),
            sort=True, ignore="IgNoRe"
        )
    if isinstance(aligned_model.latent_module, latent.SemiSupervisedCatGau):
        data_dict[supervision] = utils.encode_onehot(
            dataset.obs[supervision].astype(object).fillna("IgNoRe"),
            sort=True, ignore="IgNoRe"
        )
        cat_dim = aligned_model.latent_module.cat_dim
        if cat_dim > data_dict[supervision].shape[1]:
            data_dict[supervision] = scipy.sparse.hstack([
                data_dict[supervision].tocsc(),
                scipy.sparse.csc_matrix((
                    data_dict[supervision].shape[0],
                    cat_dim - data_dict[supervision].shape[1]
                ))
            ]).tocsr()

    aligned_model.fit(
        data_dict, batch_size=batch_size, val_split=val_split,
        epoch=epoch, patience=patience, tolerance=tolerance,
        progress_bar=progress_bar
    )
    return aligned_model



r"""
Latent space / encoder modules for DIRECTi
"""


import typing
import abc

import numpy as np
import sklearn.metrics
import tensorflow as tf

from . import module, nn, utils


class Latent(module.Module):
    r"""
    Abstract base class for latent variable modules.
    """
    def __init__(
            self, latent_dim: int, h_dim: int = 128, depth: int = 1,
            dropout: float = 0.0, lambda_reg: float = 0.0,
            fine_tune: bool = False, deviation_reg: float = 0.0,
            name: str = "Latent"
    ) -> None:
        super(Latent, self).__init__(name=name)
        self.latent_dim = latent_dim
        self.h_dim = h_dim
        self.depth = depth
        self.dropout = dropout
        self.lambda_reg = lambda_reg
        self.fine_tune = fine_tune
        self.deviation_reg = deviation_reg
        self.deviation_regularizer = \
            (lambda x: self.deviation_reg * tf.reduce_mean(tf.square(x))) \
            if self.fine_tune and self.deviation_reg > 0 else None

    @abc.abstractmethod
    def _build_latent(
            self, x: tf.Tensor, training_flag: tf.Tensor,
            scope: str = "encoder"
    ) -> tf.Tensor:  # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    def _build_regularizer(
            self, training_flag: tf.Tensor, epoch: tf.Tensor,
            scope: str = "regularizer"
    ) -> tf.Tensor:  # pragma: no cover
        raise NotImplementedError

    def __bool__(self) -> bool:
        return True

    def _get_config(self) -> typing.Mapping:
        return {
            "latent_dim": self.latent_dim,
            "h_dim": self.h_dim,
            "depth": self.depth,
            "dropout": self.dropout,
            "lambda_reg": self.lambda_reg,
            "fine_tune": self.fine_tune,
            "deviation_reg": self.deviation_reg,
            **super(Latent, self)._get_config()
        }


class Gau(Latent):
    r"""
    Build a Gaussian latent module. The Gaussian latent variable is used as
    cell embedding.

    Parameters
    ----------
    latent_dim
        Dimensionality of the latent variable.
    h_dim
        Dimensionality of the hidden layers in the encoder MLP.
    depth
        Number of hidden layers in the encoder MLP.
    dropout
        Dropout rate.
    lambda_reg
        Regularization strength on the latent variable.
    name
        Name of the module.
    """
    def __init__(
            self, latent_dim: int, h_dim: int = 128, depth: int = 1,
            dropout: float = 0.0, lambda_reg: float = 0.001,
            fine_tune: bool = False, deviation_reg: float = 0.0,
            name: str = "Gau"
    ) -> None:
        super(Gau, self).__init__(latent_dim, h_dim, depth, dropout, lambda_reg,
                                  fine_tune, deviation_reg, name)

    def _build_latent(
            self, x: tf.Tensor, training_flag: tf.Tensor,
            scope: str = "encoder"
    ) -> tf.Tensor:
        self.build_latent_scope = f"{scope}/{self.scope_safe_name}"
        with tf.variable_scope(self.build_latent_scope):
            dense_kwargs = [dict(
                deviation_regularizer=self.deviation_regularizer
            )] * self.depth
            if dense_kwargs:
                dense_kwargs[0]["weights_trainable"] = not self.fine_tune
            ptr = nn.mlp(
                x, [self.h_dim] * self.depth,
                dropout=self.dropout, batch_normalization=True,
                dense_kwargs=dense_kwargs, training_flag=training_flag
            )
            self.gau = tf.identity(nn.dense(
                ptr, self.latent_dim,
                deviation_regularizer=self.deviation_regularizer
            ), name="gau")
        self.vars_to_save += tf.get_collection(
            tf.GraphKeys.GLOBAL_VARIABLES, self.build_latent_scope)
        return self.gau

    def _build_regularizer(
            self, training_flag: tf.Tensor, epoch: tf.Tensor,
            scope: str = "discriminator"
    ) -> tf.Tensor:
        self.gaup_sampler = tf.distributions.Normal(loc=0.0, scale=1.0)
        self.build_regularizer_scope = f"{scope}/{self.scope_safe_name}"
        with tf.variable_scope(self.build_regularizer_scope, reuse=tf.AUTO_REUSE):
            self.gaup = self.gaup_sampler.sample((
                tf.shape(self.gau)[0], self.latent_dim))
            dropout = np.zeros(self.depth)
            dropout[1:] = self.dropout  # No dropout for first layer
            gau_pred = tf.sigmoid(nn.dense(nn.mlp(
                self.gau, [self.h_dim] * self.depth,
                dropout=dropout.tolist(), training_flag=training_flag
            ), 1), name="pred")
            gaup_pred = tf.sigmoid(nn.dense(nn.mlp(
                self.gaup, [self.h_dim] * self.depth,
                dropout=dropout.tolist(), training_flag=training_flag
            ), 1), name="prior_pred")
            self.gau_d_loss, self.gau_g_loss = nn.gan_loss(gaup_pred, gau_pred)

        self.vars_to_save += tf.get_collection(
            tf.GraphKeys.GLOBAL_VARIABLES, self.build_regularizer_scope)
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.gau_d_loss)
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.gau_g_loss)

        return self.lambda_reg * self.gau_g_loss

    def _compile(self, optimizer: str, lr: float) -> None:
        with tf.variable_scope(f"optimize/{self.scope_safe_name}"):
            optimizer = getattr(tf.train, optimizer)
            self.step = optimizer(lr).minimize(
                self.lambda_reg * self.gau_d_loss,
                var_list=tf.get_collection(
                    tf.GraphKeys.TRAINABLE_VARIABLES,
                    self.build_regularizer_scope
                )
            )
            tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, self.step)


class CatGau(Latent):
    r"""
    Build a double latent module, with a continuous Gaussian latent variable
    and a one-hot categorical latent variable for intrinsic clustering of
    the data. These two latent variabels are then combined into a single
    cell embedding vector.

    Parameters
    ----------
    latent_dim
        Dimensionality of the Gaussian latent variable.
    cat_dim
        Number of intrinsic clusters.
    h_dim
        Dimensionality of the hidden layers in the encoder MLP.
    depth
        Number of hidden layers in the encoder MLP.
    dropout
        Dropout rate.
    multiclass_adversarial
        Whether to use multi-class adversarial regularization on the
        Gaussian latent variable.
        Setting this to True makes each intrinsic cluster more Gaussian-like.
    cat_merge
        Whether to enable heuristic cluster merging during training.
    min_silhouette
        Minimal average silhouette score below which intrinsic clusters will be
        merged.
    patience
        Execute heuristic cluster merging under a "fast-ring" early stop
        mechanism, with early stop patience specified by this argument.
    lambda_reg
        Regularization strength on the latent variables.
    name
        Name of the module.
    """
    def __init__(
            self, latent_dim: int, cat_dim: int,
            h_dim: int = 128, depth: int = 1, dropout: float = 0.0,
            multiclass_adversarial: bool = False, cat_merge: bool = False,
            min_silhouette: float = 0.0, patience: int = 10,
            lambda_reg: float = 0.001, fine_tune: bool = False,
            deviation_reg: float = 0.0, name="CatGau"
    ) -> None:
        super(CatGau, self).__init__(latent_dim, h_dim, depth, dropout,
                                     lambda_reg, fine_tune, deviation_reg, name)
        self.cat_dim = cat_dim
        self.multiclass_adversarial = multiclass_adversarial
        self.min_silhouette = min_silhouette
        self.patience = patience
        self.cat_merge = cat_merge
        if cat_merge:
            self.on_epoch_end.append(self._cat_merge)

    def _build_latent(
            self, x: tf.Tensor, training_flag: tf.Tensor,
            scope: str = "encoder"
    ) -> tf.Tensor:
        self.build_latent_scope = f"{scope}/{self.scope_safe_name}"
        with tf.variable_scope(self.build_latent_scope):
            dense_kwargs = [dict(
                deviation_regularizer=self.deviation_regularizer
            )] * self.depth
            if dense_kwargs:  # Fix the first layer
                dense_kwargs[0]["weights_trainable"] = not self.fine_tune
            ptr = nn.mlp(
                x, [self.h_dim] * self.depth,
                dropout=self.dropout, batch_normalization=True,
                dense_kwargs=dense_kwargs,
                training_flag=training_flag
            )
            with tf.variable_scope("cat"):
                self.cat_logit = tf.identity(nn.dense(
                    ptr, self.cat_dim,
                    deviation_regularizer=self.deviation_regularizer
                ), name="cat_logit")
                self.cat = tf.nn.softmax(self.cat_logit, name="cat")
            with tf.variable_scope("gau"):
                self.gau = tf.identity(nn.dense(
                    ptr, self.latent_dim,
                    deviation_regularizer=self.deviation_regularizer
                ), name="gau")
            self.latent = self.gau + nn.dense(
                self.cat, self.latent_dim, use_bias=False,
                weights_initializer=tf.random_normal_initializer(stddev=0.1),
                deviation_regularizer=self.deviation_regularizer,
                scope="cluster_head"
            )
        self.vars_to_save += tf.get_collection(
            tf.GraphKeys.GLOBAL_VARIABLES, self.build_latent_scope)
        return self.latent

    def _build_regularizer(
            self, training_flag: tf.Tensor, epoch: tf.Tensor,
            scope: str = "discriminator"
    ) -> tf.Tensor:
        self.catp_mask = tf.get_variable(
            "catp_mask", initializer=np.ones(self.cat_dim), trainable=False)
        self.vars_to_save.append(self.catp_mask)
        self.catp_sampler = tf.distributions.Categorical(
            probs=self.catp_mask / tf.reduce_sum(self.catp_mask))
        self.gaup_sampler = tf.distributions.Normal(loc=0.0, scale=1.0)
        if self.multiclass_adversarial:
            return self._build_multiclass_regularizer(training_flag, scope)
        return self._build_binary_regularizer(training_flag, scope)

    def _build_binary_regularizer(
            self, training_flag: tf.Tensor, scope: str = "discriminator"
    ) -> tf.Tensor:
        self.build_regularizer_scope = f"{scope}/{self.scope_safe_name}"
        with tf.variable_scope(self.build_regularizer_scope, reuse=tf.AUTO_REUSE):
            with tf.variable_scope("cat"):
                self.catp = tf.one_hot(
                    self.catp_sampler.sample(tf.shape(self.cat)[0]),
                    depth=self.cat_dim
                )
                dropout = np.zeros(self.depth)
                dropout[1:] = self.dropout  # No dropout for first layer
                cat_pred = tf.sigmoid(nn.dense(nn.mlp(
                    self.cat, [self.h_dim] * self.depth,
                    dropout=dropout.tolist(), training_flag=training_flag
                ), 1), name="pred")
                catp_pred = tf.sigmoid(nn.dense(nn.mlp(
                    self.catp, [self.h_dim] * self.depth,
                    dropout=dropout.tolist(), training_flag=training_flag
                ), 1), name="prior_pred")
                self.cat_d_loss, self.cat_g_loss = \
                    nn.gan_loss(catp_pred, cat_pred)
            with tf.variable_scope("gau"):
                self.gaup = self.gaup_sampler.sample((
                    tf.shape(self.gau)[0], self.latent_dim))
                dropout = np.zeros(self.depth)
                dropout[1:] = self.dropout  # No dropout for first layer
                gau_pred = tf.sigmoid(nn.dense(nn.mlp(
                    self.gau, [self.h_dim] * self.depth,
                    dropout=dropout.tolist(), training_flag=training_flag
                ), 1), name="pred")
                gaup_pred = tf.sigmoid(nn.dense(nn.mlp(
                    self.gaup, [self.h_dim] * self.depth,
                    dropout=dropout.tolist(), training_flag=training_flag
                ), 1), name="prior_pred")
                self.gau_d_loss, self.gau_g_loss = \
                    nn.gan_loss(gaup_pred, gau_pred)

        self.vars_to_save += tf.get_collection(
            tf.GraphKeys.GLOBAL_VARIABLES, self.build_regularizer_scope)
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.cat_d_loss)
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.cat_g_loss)
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.gau_d_loss)
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.gau_g_loss)
        return self.lambda_reg * (self.cat_g_loss + self.gau_g_loss)

    def _build_multiclass_regularizer(
            self, training_flag: tf.Tensor, scope: str = "discriminator"
    ) -> tf.Tensor:
        self.build_regularizer_scope = f"{scope}/{self.scope_safe_name}"
        with tf.variable_scope(self.build_regularizer_scope, reuse=tf.AUTO_REUSE):
            with tf.variable_scope("cat"):
                self.catp = tf.one_hot(
                    self.catp_sampler.sample(tf.shape(self.cat)[0]),
                    depth=self.cat_dim
                )
                dropout = np.zeros(self.depth)
                dropout[1:] = self.dropout  # No dropout for first layer
                cat_pred = tf.sigmoid(nn.dense(nn.mlp(
                    self.cat, [self.h_dim] * self.depth,
                    dropout=dropout.tolist(), training_flag=training_flag
                ), 1), name="pred")
                catp_pred = tf.sigmoid(nn.dense(nn.mlp(
                    self.catp, [self.h_dim] * self.depth,
                    dropout=dropout.tolist(), training_flag=training_flag
                ), 1), name="prior_pred")
                self.cat_d_loss, self.cat_g_loss = \
                    nn.gan_loss(catp_pred, cat_pred)
            with tf.variable_scope("gau"):
                self.gaup = self.gaup_sampler.sample((
                    tf.shape(self.gau)[0], self.latent_dim))
                dropout = np.zeros(self.depth)
                dropout[1:] = self.dropout  # No dropout for first layer
                gau_logits = nn.dense(nn.mlp(
                    self.gau, [self.h_dim] * self.depth,
                    dropout=dropout.tolist(), training_flag=training_flag
                ), self.cat_dim + 1)
                gaup_logits = nn.dense(nn.mlp(
                    self.gaup, [self.h_dim] * self.depth,
                    dropout=dropout.tolist(), training_flag=training_flag
                ), self.cat_dim + 1)
                true = tf.concat([
                    tf.concat([
                        self.cat,
                        tf.zeros((tf.shape(self.cat)[0], 1))
                    ], axis=1),
                    tf.concat([
                        tf.zeros((tf.shape(gaup_logits)[0], self.cat_dim)),
                        tf.ones((tf.shape(gaup_logits)[0], 1))
                    ], axis=1)
                ], axis=0)
                logits = tf.concat([gau_logits, gaup_logits], axis=0)
                self.gau_d_loss = tf.reduce_mean(
                    tf.nn.softmax_cross_entropy_with_logits_v2(
                        labels=tf.stop_gradient(true), logits=logits,
                    ), name="d_loss"
                )
                self.gau_g_loss = tf.negative(self.gau_d_loss, name="g_loss")

        self.vars_to_save += tf.get_collection(
            tf.GraphKeys.GLOBAL_VARIABLES, self.build_regularizer_scope)
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.cat_d_loss)
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.cat_g_loss)
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.gau_d_loss)
        tf.add_to_collection(tf.GraphKeys.LOSSES, self.gau_g_loss)
        return self.lambda_reg * (self.cat_g_loss + self.gau_g_loss)

    def _compile(self, optimizer: str, lr: float) -> None:
        with tf.variable_scope(f"optimize/{self.scope_safe_name}"):
            optimizer = getattr(tf.train, optimizer)
            self.step = optimizer(lr).minimize(
                self.lambda_reg * (self.cat_d_loss + self.gau_d_loss),
                var_list=tf.get_collection(
                    tf.GraphKeys.TRAINABLE_VARIABLES,
                    self.build_regularizer_scope
                )
            )
            tf.add_to_collection(tf.GraphKeys.UPDATE_OPS, self.step)

    # On epoch end heuristic
    def _cat_merge(
            self, model: "directi.DIRECTi",
            train_data_dict: utils.DataDict,
            val_data_dict: utils.DataDict,  # pylint: disable=unused-argument
            loss: tf.Tensor
    ) -> bool:

        # Initialization
        if "_cat_merge_dict" not in dir(self):
            self._cat_merge_dict = {
                "record": np.inf,
                "countdown": self.patience,
                "converged": False
            }
        d = self._cat_merge_dict

        # Guard operatability
        if model.sess.run(self.catp_mask).sum() == 1:
            d["converged"] = True
            return d["converged"]

        # Guard entrance
        if loss < d["record"]:
            d["record"] = loss
            d["countdown"] = self.patience
        else:
            d["countdown"] -= 1

        # Entrance
        if d["countdown"] == 0:
            removed_clusters = set(np.where(
                model.sess.run(self.catp_mask) == 0
            )[0])

            # Identify cluster heads that are not assigned any samples
            cluster = model._fetch(
                self.cat, train_data_dict
            ).argmax(axis=1).astype(np.int)
            population = np.eye(self.cat_dim)[cluster, :].sum(axis=0)
            remove_idx = set(np.where(
                population <= 1
            )[0]).difference(removed_clusters)
            reason = "emptyness"

            # Identify clusters that are not clearly separated
            if not remove_idx:
                latent = model._fetch(self.latent, train_data_dict)
                if train_data_dict.size > 10000:
                    subsample_idx = model.random_state.choice(
                        train_data_dict.size, 10000, replace=False)
                    latent = latent[subsample_idx]
                    cluster = cluster[subsample_idx]
                    population = np.eye(self.cat_dim)[cluster, :].sum(axis=0)
                sample_silhouette = sklearn.metrics.silhouette_samples(
                    latent, cluster)
                cluster_silhouette = np.empty(len(population))
                for i, _population in enumerate(population):
                    cluster_silhouette[i] = sample_silhouette[
                        cluster == i
                    ].mean() if _population > 0 else np.inf
                population[cluster_silhouette > self.min_silhouette] = np.inf
                remove_candidate = set(
                    np.where(np.isfinite(population))[0]
                ).difference(removed_clusters).difference(self._safe_cat(model))
                if remove_candidate:
                    remove_candidate = list(remove_candidate)
                    remove_idx = {remove_candidate[
                        population[remove_candidate].argmin()
                    ]}
                    reason = "inconfidence"
                else:
                    remove_idx = dict()

            # Merge identified clusters
            if remove_idx:
                d["converged"] = False
                current_value = model.sess.run(self.catp_mask)
                current_value[list(remove_idx)] = 0
                model.sess.run(tf.assign(self.catp_mask, current_value))
                print("Cluster %s removed because of %s." % (
                    ",".join([str(item) for item in remove_idx]), reason
                ))
            else:
                d["converged"] = True
                print("Nothing done in cluster manipulation.")

            d["countdown"] = self.patience
            d["record"] = np.inf

        return d["converged"]

    def _safe_cat(
            self, model: "directi.DIRECTi"  # pylint: disable=unused-argument
    ) -> typing.List:  # define which clusters should safe from merging
        return []

    def _get_config(self) -> typing.Mapping:
        return {
            "cat_dim": self.cat_dim,
            "multiclass_adversarial": self.multiclass_adversarial,
            "cat_merge": self.cat_merge,
            "min_silhouette": self.min_silhouette,
            "patience": self.patience,
            **super(CatGau, self)._get_config()
        }


class SemiSupervisedCatGau(CatGau):
    r"""
    Build a double latent module, with a continuous Gaussian latent variable
    and a one-hot categorical latent variable for intrinsic clustering of
    the data. The categorical latent supports semi-supervision. The two latent
    variables are then combined into a single cell embedding vector.

    Parameters
    ----------
    latent_dim
        Dimensionality of the Gaussian latent variable.
    cat_dim
        Number of intrinsic clusters.
    h_dim
        Dimensionality of the hidden layers in the encoder MLP.
    depth
        Number of hidden layers in the encoder MLP.
    dropout
        Dropout rate.
    multiclass_adversarial
        Whether to use multi-class adversarial regularization on the
        Gaussian latent variable.
        Setting this to True makes each intrinsic cluster more Gaussian-like.
    cat_merge
        Whether to enable heuristic cluster merging during training.
    min_silhouette
        Minimal average silhouette score required to prevent an instrinsic
        cluster from being merged.
    patience
        Execute heuristic cluster merging under a "fast-ring" early stop
        mechanism, with early stop patience specified by this argument.
    lambda_sup
        Supervision strength.
    background_catp
        Unnormalized background prior distribution of the intrinsic
        clustering latent.
        For each supervised cell in a minibatch, unnormalized prior
        probability of the corresponding cluster will increase by 1,
        so this parameter determines how much to trust supervision class
        frequency, and it balances between supervision and identifying new
        clusters.
    lambda_reg
        Regularization strength on the latent variables.
    name
        Name of latent module.
    """
    def __init__(
            self, latent_dim: int, cat_dim: int,
            h_dim: int = 128, depth: int = 1, dropout: float = 0.0,
            multiclass_adversarial: bool = False, cat_merge: bool = False,
            min_silhouette: float = 0.0, patience: int = 10,
            lambda_sup: float = 10.0, background_catp: float = 1e-3,
            lambda_reg: float = 0.001, fine_tune: bool = False,
            deviation_reg: float = 0.0, name: str = "SemiSupervisedCatGau"
    ) -> None:
        super(SemiSupervisedCatGau, self).__init__(
            latent_dim, cat_dim, h_dim, depth, dropout, multiclass_adversarial,
            cat_merge, min_silhouette, patience, lambda_reg,
            fine_tune, deviation_reg, name
        )
        self.lambda_sup = lambda_sup
        self.background_catp = background_catp

    def _build_regularizer(
            self, training_flag: tf.Tensor, epoch: tf.Tensor,
            scope: str = "discriminator"
    ) -> tf.Tensor:
        with tf.name_scope("placeholder/"):
            self.cats = tf.placeholder(
                dtype=tf.float32, shape=(None, self.cat_dim), name="cats")
        self.cats_coverage = tf.get_variable(
            "cats_coverage", initializer=np.zeros(
                self.cat_dim, dtype=np.bool_
            ), trainable=False
        )  # Used in cat manipulation to avoid merging supervised label
        self.vars_to_save.append(self.cats_coverage)

        self.catp_mask = tf.get_variable(
            "catp_mask", initializer=np.ones(
                self.cat_dim, dtype=np.float32
            ) * self.background_catp, trainable=False
        )
        self.catp_prob = self.catp_mask + tf.reduce_sum(self.cats, axis=0)
        self.catp_sampler = tf.distributions.Categorical(
            probs=self.catp_prob / tf.reduce_sum(self.catp_prob))
        self.gaup_sampler = tf.distributions.Normal(loc=0.0, scale=1.0)
        self.vars_to_save.append(self.catp_mask)

        with tf.name_scope(f"semi_supervision/{self.scope_safe_name}"):
            mask = tf.cast(tf.reduce_sum(self.cats, axis=1) > 0, tf.int32)
            masked_cat_logit = tf.dynamic_partition(self.cat_logit, mask, 2)[1]
            masked_cats = tf.dynamic_partition(self.cats, mask, 2)[1]
            self.sup_loss = tf.reduce_mean(
                tf.nn.softmax_cross_entropy_with_logits_v2(
                    labels=masked_cats, logits=masked_cat_logit
                ), name="supervised_loss"
            )
            tf.add_to_collection(tf.GraphKeys.LOSSES, self.sup_loss)

        tf.add_to_collection(
            tf.GraphKeys.UPDATE_OPS,
            tf.assign(self.cats_coverage, tf.logical_or(
                self.cats_coverage,
                tf.cast(tf.reduce_max(self.cats, axis=0), tf.bool)
            ))
        )
        return self.lambda_sup * self.sup_loss + (
            self._build_multiclass_regularizer
            if self.multiclass_adversarial
            else self._build_binary_regularizer
        )(training_flag, scope)

    def _build_feed_dict(self, data_dict: utils.DataDict) -> typing.Mapping:
        return {
            self.cats: utils.densify(data_dict[self.name])
        } if self.name in data_dict else {}

    def _safe_cat(self, model: "directi.DIRECTi") -> typing.List:
        return np.where(model.sess.run(self.cats_coverage))[0]

    def _get_config(self) -> typing.Mapping:
        return {
            "lambda_sup": self.lambda_sup,
            "background_catp": self.background_catp,
            **super(SemiSupervisedCatGau, self)._get_config()
        }



r"""
Neural network operations
"""

import typing

import tensorflow as tf
from tensorflow.python.ops.init_ops import Initializer  # pylint: disable=no-name-in-module


def mlp(
        input_tensor: typing.Union[tf.Tensor, typing.List[tf.Tensor]],
        hidden_dim: int, activation: typing.Union[
            typing.Callable[[tf.Tensor], tf.Tensor],
            typing.List[typing.Callable[[tf.Tensor], tf.Tensor]]
        ] = tf.nn.leaky_relu,
        dropout: typing.Union[float, typing.List[float]] = 0.0,
        batch_normalization: typing.Union[bool, typing.List[bool]] = False,
        dense_kwargs: typing.Union[
            typing.Mapping, typing.List[typing.Mapping]
        ] = None,
        training_flag: typing.Optional[tf.Tensor] = None, scope: str = "mlp"
) -> tf.Tensor:
    if not isinstance(hidden_dim, list):
        hidden_dim = [hidden_dim]
    if isinstance(activation, list):
        assert len(activation) == len(hidden_dim)
    else:
        activation = [activation] * len(hidden_dim)
    if isinstance(dropout, list):
        assert len(dropout) == len(hidden_dim)
    else:
        dropout = [dropout] * len(hidden_dim)
    if isinstance(batch_normalization, list):
        assert len(batch_normalization) == len(hidden_dim)
    else:
        batch_normalization = [batch_normalization] * len(hidden_dim)
    if isinstance(dense_kwargs, list):
        assert len(dense_kwargs) == len(hidden_dim)
    else:
        dense_kwargs = [dense_kwargs] * len(hidden_dim)

    with tf.variable_scope(scope):
        ptr = input_tensor
        for l, (_hidden_dim, _activation, _dropout, _batch_normalization, _dense_kwargs) in enumerate(zip(
                hidden_dim, activation, dropout, batch_normalization, dense_kwargs
        )):
            assert not (_dropout or _batch_normalization) or training_flag is not None
            with tf.variable_scope(f"layer_{l}"):
                ptr = tf.layers.dropout(
                    ptr, rate=_dropout, training=training_flag
                ) if _dropout else ptr
                this_dense_kwargs = dict(
                    output_dim=_hidden_dim,
                    use_bias=not _batch_normalization
                )
                if _dense_kwargs is not None:
                    this_dense_kwargs.update(_dense_kwargs)
                ptr = dense(ptr, **this_dense_kwargs)
                ptr = tf.layers.batch_normalization(
                    ptr, center=True, scale=True, training=training_flag
                ) if _batch_normalization else ptr
                ptr = _activation(ptr) if _activation is not None else ptr
    return ptr


def dense(
        input_tensor: typing.Union[tf.Tensor, typing.List[tf.Tensor]],
        output_dim: int, use_bias: bool = True,
        weights_initializer: Initializer = tf.truncated_normal_initializer(stddev=0.01),
        bias_initializer: Initializer = tf.zeros_initializer(),
        weights_trainable: bool = True, bias_trainable: bool = True,
        weights_regularizer: typing.Optional[
            typing.Callable[[tf.Tensor], tf.Tensor]
        ] = None,
        bias_regularizer: typing.Optional[
            typing.Callable[[tf.Tensor], tf.Tensor]
        ] = None,
        deviation_regularizer: typing.Optional[
            typing.Callable[[tf.Tensor], tf.Tensor]
        ] = None, scope: str = "dense"
) -> tf.Tensor:
    # `input_tensor` can accept a list of tensors instead of concatenating
    # them before passing to `dense`, which enables reusing part of the weight
    # matrix later.
    # Similarly, `weights_initializer`, `weights_trainable` and
    # `weights_regularizer` can also be a list, each element corresponding to
    # weights of an input tensor.
    with tf.variable_scope(scope):
        if not isinstance(input_tensor, list):
            input_tensor = [input_tensor]
        if isinstance(weights_initializer, list):
            assert len(weights_initializer) == len(input_tensor)
        else:
            weights_initializer = [weights_initializer] * len(input_tensor)
        if isinstance(weights_trainable, list):
            assert len(weights_trainable) == len(input_tensor)
        else:
            weights_trainable = [weights_trainable] * len(input_tensor)
        if isinstance(weights_regularizer, list):
            assert len(weights_regularizer) == len(input_tensor)
        else:
            weights_regularizer = [weights_regularizer] * len(input_tensor)
        ptr = []
        for i, (
                _input_tensor, _weights_initializer,
                _weights_trainable, _weights_regularizer
        ) in enumerate(zip(
            input_tensor, weights_initializer,
            weights_trainable, weights_regularizer
        )):
            input_dim = _input_tensor.get_shape().as_list()[1]
            name = f"weights_{i}" if i > 0 else "weights"
            weights = tf.get_variable(
                name, shape=(input_dim, output_dim), dtype=tf.float32,
                initializer=_weights_initializer if deviation_regularizer is None
                else tf.zeros_initializer(),
                trainable=_weights_trainable
            )
            if _weights_regularizer is not None:
                reg = _weights_regularizer(weights)
                tf.add_to_collection(tf.GraphKeys.REGULARIZATION_LOSSES, reg)
            if deviation_regularizer is not None:
                name += "_ori"
                weights_ori = tf.get_variable(
                    name, shape=(input_dim, output_dim),
                    dtype=tf.float32, trainable=False
                )
                assign = tf.assign(weights_ori, weights)
                reg = deviation_regularizer(weights - weights_ori)
                tf.add_to_collection(tf.GraphKeys.READY_OP, assign)
                tf.add_to_collection(tf.GraphKeys.REGULARIZATION_LOSSES, reg)
            ptr.append(tf.matmul(_input_tensor, weights))
        ptr = tf.add_n(ptr)
        if use_bias:
            name = "bias"
            bias = tf.get_variable(
                name, shape=(output_dim, ), dtype=tf.float32,
                initializer=bias_initializer, trainable=bias_trainable
            )
            if bias_regularizer is not None:
                reg = bias_regularizer(bias)
                tf.add_to_collection(tf.GraphKeys.REGULARIZATION_LOSSES, reg)
            if deviation_regularizer is not None:
                name += "_ori"
                bias_ori = tf.get_variable(
                    name, shape=(output_dim, ), dtype=tf.float32,
                    initializer=tf.zeros_initializer(), trainable=False
                )
                assign = tf.assign(bias_ori, bias)
                reg = deviation_regularizer(bias - bias_ori)
                tf.add_to_collection(tf.GraphKeys.READY_OP, assign)
                tf.add_to_collection(tf.GraphKeys.REGULARIZATION_LOSSES, reg)
            ptr = tf.add(ptr, bias)
    return ptr


def gan_loss(
        true: tf.Tensor, fake: tf.Tensor, eps: float = 1e-8
) -> typing.Tuple[tf.Tensor, tf.Tensor]:
    with tf.name_scope("d_loss"):
        d_loss = tf.negative(tf.reduce_mean(
            tf.log(true + eps) + tf.log(1. - fake + eps)
        ), name="d_loss")
    with tf.name_scope("g_loss"):
        g_loss = tf.negative(tf.reduce_mean(
            tf.log(fake + eps)
        ), name="g_loss")
    return d_loss, g_loss
